"""Contains tools needed to run the command line interface."""
import sys
import argparse
from antpack import SingleChainAnnotator, PairedChainAnnotator, VJGeneTool

from .cli_tools.biofile_read_tools import read_fasta


class ReconfigParser(argparse.ArgumentParser):
    """Reconfigure argparse's parser so an automatic error message
    is generated if no args supplied."""
    def error(self, message):
        self.print_help()
        sys.exit(2)


def gen_arg_parser():
    """Build the command line arg parser."""
    parser = ReconfigParser(description="Run AntPack from the command line "
            "to number and provide VJ assignments for input AA sequences. "
            "This command line tool first reads all input sequences into memory, "
            "and writes them to a csv in a format that makes them easy to review "
            "and compare (appropriate for small-moderate datasets). If you need to work "
            "with a larger dataset or do a more customized or extensive analysis, "
            "use the Python API instead.")
    parser.add_argument("input", nargs = 1, help=
            "Input filepath. Must be in fasta format.")
    parser.add_argument("output", nargs = 1, help=
            "Output filepath. Two output files are created, one output "
            "file with '_heavy.csv' appended to the output path for "
            "heavy chains (if any) and one with '_light.csv' appended "
            "to the output path for light chains (if any).")
    parser.add_argument("scheme", nargs = 1, help=
            "The numbering scheme. One of aho, imgt, kabat or martin.")
    parser.add_argument("--paired", action="store_true", help=
            "AntPack will normally assume there is one variable region "
            "per input sequence and try to extract it. If this "
            "flag is supplied, it will instead assume each input "
            "sequence contains a heavy chain and a light chain and "
            "try to extract both. The --paired flag can still be "
            "supplied if some of the sequences are single-chain.")
    parser.add_argument("--vj", nargs=2, help=
            "If this flag is supplied, AntPack will find the closest "
            "VJ genes for the specified species and write these "
            "into the output. There are two arguments. The first is "
            "the species which must be either human or mouse. The "
            "second is the mode which must be identity or evalue. ",
            metavar=("species", "mode"))
    parser.add_argument("--chains", nargs=1, help=
            "AntPack will normally look for an H, K or L chain in each "
            "input sequence. If desired, you can instead restrict it to "
            "search for a comma-separated list of specific chains (e.g. H or "
            "K,L) by using this argument. If you supply the argument "
            "--paired this argument is ignored.",
            metavar=("chains"))
    return parser






def process_fasta_online(cli_args, chains):
    """Processes a fasta file by first loading all sequences
    into memory. This is faster and simpler but obviously
    greatly increases memory consumption if the number of
    sequences in the input file is large. For a large
    file, it is strongly recommended to use the Python
    API instead."""
    seqrecs = list(read_fasta(cli_args.input[0]))
    seqs = [seqrec[1] for seqrec in seqrecs]
    seqinfo = [seqrec[0] for seqrec in seqrecs]
    vj_tool = VJGeneTool(scheme = cli_args.scheme[0])

    output_dict = {
        k:{"annotations":[], "seqs":[],
        "seqinfo":[], "v_genes":[], "j_genes":[],
        "v_scores":[], "j_scores":[], "msa":None}
        for k in ["heavy", "light"]
        }

    if cli_args.paired:
        sc_tool = PairedChainAnnotator(cli_args.scheme[0])
        seq_annotations = [sc_tool.analyze_seq(s) for s in seqs]
        for i, chain in enumerate(["heavy", "light"]):
            output_dict[chain]["annotations"] = [s[i] for s in seq_annotations]
            output_dict[chain]["seqs"] = seqs
            output_dict[chain]["seqinfo"] = seqinfo
            output_dict[chain]["msa"] = sc_tool.build_msa(seqs,
                    output_dict[chain]["annotations"], True)

        del seq_annotations

    else:
        sc_tool = SingleChainAnnotator(chains, cli_args.scheme[0])
        annotations = sc_tool.analyze_seqs(seqs)

        for i, (seq, annot) in enumerate(zip(seqs, annotations)):
            if annot[2] in ("K", "L"):
                output_dict["light"]["annotations"].append(annot)
                output_dict["light"]["seqs"].append(seq)
                output_dict["light"]["seqinfo"].append(seqinfo[i])
            else:
                output_dict["heavy"]["annotations"].append(annot)
                output_dict["heavy"]["seqs"].append(seq)
                output_dict["heavy"]["seqinfo"].append(seqinfo[i])

        for chain in ["heavy", "light"]:
            if len(output_dict[chain]["seqs"]) > 0:
                output_dict[chain]["msa"] = sc_tool.build_msa(output_dict[chain]["seqs"],
                        output_dict[chain]["annotations"], True)

    if cli_args.vj is not None:
        for chain in ["heavy", "light"]:
            for seq, annot in zip(output_dict[chain]["seqs"], output_dict[chain]["annotations"]):
                v_gene, j_gene, v_ident, j_ident = vj_tool.assign_vj_genes(annot, seq,
                        cli_args.vj[0], cli_args.vj[1])
                output_dict[chain]["v_genes"].append(v_gene)
                output_dict[chain]["j_genes"].append(j_gene)
                output_dict[chain]["v_scores"].append(v_ident)
                output_dict[chain]["j_scores"].append(j_ident)

    for chain, cdict in output_dict.items():
        if cdict["msa"] is None:
            continue
        with open(cli_args.output[0] + f"_{chain}.csv", "w+",
                encoding="utf-8") as fhandle:
            if cli_args.vj is not None:
                _ = fhandle.write("Sequence_info,percent_identity,vj_species,"
                        "vj_mode,v_gene,v_score,j_gene,j_score,")
            else:
                _ = fhandle.write("Sequence_info,percent_identity,")
            _ = fhandle.write(f"{','.join(cdict['msa'][0])},error_message\n")

            if not cli_args.vj:
                for i, (seq_id, annotation, msa_row) in enumerate(zip(cdict["seqinfo"],
                    cdict["annotations"], cdict["msa"][1])):
                    percent_identity = 100 * round(annotation[1], 3)
                    _ = fhandle.write(f"{seq_id},{percent_identity},")
                    _ = fhandle.write(",".join(list(msa_row)))
                    _ = fhandle.write(f",{cdict['annotations'][i][3]}\n")

            else:
                for i, (seq_id, annotation, msa_row) in enumerate(zip(cdict["seqinfo"],
                    cdict["annotations"], cdict["msa"][1])):
                    percent_identity = 100 * round(annotation[1], 3)
                    _ = fhandle.write(f"{seq_id},{percent_identity},")
                    _ = fhandle.write(",".join([cli_args.vj[0], cli_args.vj[1],
                        cdict["v_genes"][i], str(cdict["v_scores"][i]),
                        cdict["j_genes"][i], str(cdict["j_scores"][i]) ] +
                        list(msa_row) ))
                    _ = fhandle.write(f",{cdict['annotations'][i][3]}\n")



def run_cli_interface():
    """Runs the command line interface using the user's specified
    options."""
    parser = gen_arg_parser()
    if len(sys.argv) < 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    if args.chains:
        chains = args.chains.split(",")
    else:
        chains = ['H', 'K', 'L']

    process_fasta_online(args, chains)



"""Tests basic functionality for the LiabilitySearchTool class."""
import os
import unittest
from antpack import LiabilitySearchTool

class TestLiabilitySearchTool(unittest.TestCase):


    def test_liability_tool(self):
        """Check that sequences with liabilities are flagged as such
        while other sequences are passed."""
        search_tool = LiabilitySearchTool()

        dummy_sequence = "AATHSCSSC"
        result = search_tool.analyze_seq(dummy_sequence)
        self.assertTrue(result[0][1] == "Sequence contains invalid characters")

        generic_test_seq = "VHLQQSGAELMKPGASVKISCKASGYTFITYWIEWVKQRPGHGLEWIGDILPGSGSTNYNENFKGKATFTADSSSNTAYMQLSSLTSEDSAVYYCARSGYYGNSGFAYWGQGTLVTVSA"
        result = search_tool.analyze_seq(generic_test_seq)
        self.assertTrue(len(result)==3)
        self.assertTrue(result[0][1] == "Deamidation (elevated risk)")
        self.assertTrue(result[1][1] == "Methionine / Tryptophan oxidation moderate risk")
        self.assertTrue(result[2][1] == "Deamidation (low risk)")

        corrected_seq = list(generic_test_seq)
        corrected_seq[102:104] = "AL"
        self.assertTrue(len(search_tool.analyze_seq("".join(corrected_seq))) == 2)
        corrected_seq[31] = "A"
        self.assertTrue(len(search_tool.analyze_seq("".join(corrected_seq))) == 1)
        corrected_seq[56:58] = "AL"
        self.assertTrue(len(search_tool.analyze_seq("".join(corrected_seq))) == 0)

        liability_seq = corrected_seq.copy()
        liability_seq[3] = "C"
        result = search_tool.analyze_seq("".join(liability_seq))
        self.assertTrue(result[0][1] == "Unusual cysteine")

        liability_seq = corrected_seq.copy()
        liability_seq[-5:-2] = "NAS"
        result = search_tool.analyze_seq("".join(liability_seq))
        self.assertTrue(result[0][1] == "N-glycosylation")

        liability_seq = corrected_seq.copy()
        liability_seq[3:5] = "M"
        result = search_tool.analyze_seq("".join(liability_seq))
        self.assertTrue(len(result) == 0)


if __name__ == "__main__":
    unittest.main()



"""Implements the CategoricalMixture class, for all operations involved
in generating mixture model predictions for new datapoints. The tools
necessary to fit the model (via EM) are in a separate repo.
Fitting right now is strictly via EM although we may evaluate
variational methods in future."""
import numpy as np
from ..utilities.special_functions import logsumexp
from .scoring_constants import catmix_constants as constants
from antpack.antpack_cpp_ext import getProbsCExt, getProbsCExt_masked, mask_terminal_deletions



class CategoricalMixture:
    """A CategoricalMixture model, with all the methods necessary
    to score and do inference.

    Attributes:
        mu_mix (np.ndarray): Array of type np.float64, shape (self.n_components,
            self.sequence_length, self.num_possible_items). The probability of
            each possible item for each point in sequence length for each cluster.
            Initialized to None, only converted to array once model is fitted.
        log_mu_mix (np.ndarray): Array of type np.float64, shape (self.n_components,
            self.sequence_length, self.num_possible_items). The log-probability of
            each possible item for each point in sequence length for each cluster.
            Initialized to None, only converted to array once model is fitted.
        mix_weights (np.ndarray): Array of type np.float64, shape (self.n_components).
            The weight for each distribution in the mixture.
        log_mix_weights (np.ndarray): Array of type np.float64, shape (self.n_components).
            The log of the mixture weight for each distribution in the mixture.
        n_components (int): The number of components.
        num_possible_items (int): The number of possible choices
            at each position in the sequence.
        sequence_length (int): The length of the sequences that the
            model will be fitted to / can analyze.
            
    """

    def __init__(self, n_components, num_possible_items = 21,
                sequence_length = 158):
        """Class constructor.

        Args:
            n_components (int): The number of mixture components
                (i.e. number of clusters).
            num_possible_items (int): The number of possible choices
                at each position in the sequence. For a protein sequence,
                for example, this might be the number of amino acid symbols
                that are possible; for a sequence of letters, this might
                be the number of letters in the alphabet; for shopping
                data, this might be the number of unique items the customer
                might purchase. Currently limited to the range from 1 - 255,
                this restriction will be lifted in a future version.
            sequence_length (int): The length of the sequences that the
                model will be fitted to / can analyze.

        Raises:
            ValueError: A ValueError is raised if unacceptable arguments are
                supplied.
        """
        if n_components <= 0:
            raise ValueError("n_components must be > 0.")
        if num_possible_items > 255 or num_possible_items <= 0:
            raise ValueError("Currently num_possible_items is limited to "
                    "values from 1 to 255, inclusive.")
        if sequence_length <= 0:
            raise ValueError("Sequence length must be positive.")

        self.mix_weights = None
        self.log_mix_weights = None
        self.mu_mix = None
        self.log_mu_mix = None
        self.n_components = n_components
        self.num_possible_items = num_possible_items
        self.sequence_length = sequence_length


    def load_params(self, mu_mix, mix_weights):
        """Checks input params to ensure they are compatible with selected
        settings, and if so, sets the model parameters to the inputs.

        Args:
            mu_mix (np.ndarray): Array of type np.float64, shape (self.n_components,
                self.sequence_length, self.num_possible_items). The probability of
                each possible item for each point in sequence length for each cluster.
            mix_weights (np.ndarray): Array of type np.float64, shape (self.n_components).
                The weight for each distribution in the mixture.

        Raises:
            ValueError: A ValueError is raised if the inputs are inappropriate.
        """
        if not isinstance(mu_mix, np.ndarray) or not isinstance(mix_weights, np.ndarray):
            raise ValueError("mu_mix and mix_weights should both be numpy arrays.")
        if len(mu_mix.shape) != 3 or len(mix_weights.shape) != 1:
            raise ValueError("mu_mix and mix_weights must be 3d and 1d arrays respectively.")
        if mu_mix.shape[0] != self.n_components or mu_mix.shape[1] != self.sequence_length \
                or mu_mix.shape[2] != self.num_possible_items:
            raise ValueError("mu_mix has an inappropriate shape.")
        if mix_weights.shape[0] != self.n_components:
            raise ValueError("mix_weights has an inappropriate shape.")

        if not mu_mix.flags["C_CONTIGUOUS"] or not mix_weights.flags["C_CONTIGUOUS"]:
            raise ValueError("Non-C-contiguous arrays supplied.")

        self.mix_weights = mix_weights
        self.log_mix_weights = mix_weights.copy()
        self.log_mix_weights[self.log_mix_weights < constants.MINIMUM_PROB_VAL] = constants.MINIMUM_PROB_VAL
        self.log_mix_weights = np.log(self.log_mix_weights)

        self.mu_mix = mu_mix
        self.log_mu_mix = mu_mix.copy()
        self.log_mu_mix[self.log_mu_mix < constants.MINIMUM_PROB_VAL] = constants.MINIMUM_PROB_VAL
        self.log_mu_mix[:] = np.log(self.log_mu_mix)



    def _check_input_array(self, xdata, n_threads):
        """Checks an input array to make sure that all have
        the correct format to ensure that no problems will be encountered.

        Args:
            x (np.ndarray): A numpy array with the input data.
            n_threads (int): The number of threads requested.

        Raises:
            ValueError: A ValueError is raised if unacceptable
                input data is supplied.
        """
        if not isinstance(xdata, np.ndarray):
            raise ValueError("Unexpected input supplied.")
        if np.max(xdata) >= self.num_possible_items or np.min(xdata) < 0:
            raise ValueError("Values in input data are out of range.")
        if xdata.dtype != "uint8":
            raise ValueError("Unexpected datatype for input data.")
        if len(xdata.shape) != 2:
            raise ValueError("Unexpected shape for input data.")
        if xdata.shape[1] != self.sequence_length or xdata.shape[0] < 1:
            raise ValueError("Unexpected shape for input data.")
        if not xdata.flags["C_CONTIGUOUS"]:
            raise ValueError("Input data is not C-contiguous.")

        if n_threads <= 0:
            raise ValueError("n_threads should be a positive integer.")
        if not isinstance(n_threads, int):
            raise ValueError("n_threads should be a positive integer.")


    def prep_masked_array(self, xdata, mask = None, mask_terminal_dels = False,
            mask_gaps = False):
        """Prepares a masked version of an input array if any masking options are
        supplied.

        Args:
            xdata (np.ndarray): A 2d numpy array of type np.uint8.
            mask (np.ndarray): Either None or a numpy array of type bool
                and shape (xdata.shape[1]). If not None, indicated
                positions are masked, i.e. are not taken into account
                when calculating the score.
            mask_gaps (bool): If True, all non-filled IMGT positions in the sequence
                are ignored when calculating the score. This is useful when your
                sequence has unusual deletions and you would like to ignore these.
            mask_terminal_dels (bool): If True, ignore N- and C-terminal
                deletions when assigning to a cluster.

        Returns:
            xmasked (np.ndarray): A 2d numpy array of type np.uint8.

        Raises:
            ValueError: A ValueError is raised if bad inputs are supplied.
        """
        xmasked = xdata.copy()

        if mask is not None:
            if not isinstance(mask, np.ndarray):
                raise ValueError("Mask must be a numpy array.")
            if mask.shape[0] != xdata.shape[1]:
                raise ValueError("Mask shape must be consistent with xdata shape.")
            if len(mask.shape) != 1:
                raise ValueError("Mask must be a 1d array.")

            #There is no amino acid 21, so we use this as a convenient "please ignore"
            #indicator.
            xmasked[:,~mask] = 21

        if mask_terminal_dels:
            mask_terminal_deletions(xmasked)

        if mask_gaps:
            xmasked[xmasked==20] = 21

        return xmasked



    def predict(self, xdata, mask = None, mask_terminal_dels = False,
            mask_gaps = False, use_mixweights = True, return_raw_probs = False,
            n_threads = 1):
        """Determine the most probable cluster for each datapoint
        in a numpy array. Note that you should also check the
        overall probability of each datapoint. If a datapoint is
        very different from your training set, it will have
        very low overall probability, but this function will
        still assign it to the most likely cluster -- whichever
        that is -- by default.

        Args:
            xdata (np.ndarray): A 2d numpy array of type np.uint8.
            n_threads (int): The number of threads to use.
            mask (np.ndarray): Either None or a numpy array of type bool
                and shape (xdata.shape[1]). If not None, indicated
                positions are masked, i.e. are not taken into account
                when calculating the score.
            mask_gaps (bool): If True, all non-filled IMGT positions in the sequence
                are ignored when calculating the score. This is useful when your
                sequence has unusual deletions and you would like to ignore these.
            mask_terminal_dels (bool): If True, ignore N- and C-terminal
                deletions when assigning to a cluster.
            use_mixweights (bool): If True, take mixture weights into
                account; otherwise, find the closest cluster (even if
                it is a low-probability cluster).
            return_raw_probs (bool): If True, return the raw probabilities
                instead of cluster assignments.

        Returns:
            preds (np.ndarray): EITHER an array of shape (xdata.shape[0])
                containing a number from 0 to self.n_components - 1
                indicating the predicted cluster for each datapoint if
                return_raw_probs is False, OR an array of shape (ncomponents,
                xdata.shape[0]) if return_raw_probs is True, indicating
                the probability of each cluster for each datapoint.

        Raises:
            ValueError: Raised if unexpected inputs are supplied.
        """
        self._check_input_array(xdata, n_threads)
        if self.log_mu_mix is None or self.log_mix_weights is None:
            raise ValueError("Model not fitted yet.")

        resp = np.zeros((self.log_mu_mix.shape[0], xdata.shape[0]))

        if mask is not None or mask_terminal_dels or mask_gaps:
            xmasked = self.prep_masked_array(xdata, mask, mask_terminal_dels, mask_gaps)
            getProbsCExt_masked(xmasked, self.log_mu_mix, resp, n_threads)

        else:
            getProbsCExt(xdata, self.log_mu_mix, resp, n_threads)

        if use_mixweights:
            resp += self.log_mix_weights[:,None]
        if return_raw_probs:
            return resp
        cluster_assignments = resp.argmax(axis=0).astype(np.uint32)
        return cluster_assignments



    def score(self, xdata, mask = None, mask_terminal_dels = False,
            mask_gaps = False, normalize_scores = False, n_threads = 1):
        """Generate the overall log-likelihood of individual datapoints.
        This is very useful to determine if a new datapoint is very
        different from the training set. If the log-likelihood of
        the new datapoints is much lower than the training set
        distribution of log-likelihoods, it is fairly unlikely
        that the new datapoint is a sample from the distribution
        represented by the model, and you should not try to
        assign it to a cluster.

        Args:
            xdata (np.ndarray): An array with the input data,
                of type np.uint8.
            mask (np.ndarray): Either None or a numpy array of type bool
                and shape (xdata.shape[1]). If not None, indicated
                positions are masked, i.e. are not taken into account
                when calculating the score.
            mask_terminal_dels (bool): If True, ignore N- and C-terminal
                deletions when calculating a score.
            mask_gaps (bool): If True, all non-filled IMGT positions in the sequence
                are ignored when calculating the score. This is useful when your
                sequence has unusual deletions and you would like to ignore these.
            normalize_scores (bool): If True, normalize the score by dividing by
                the number of non-masked residues in the input.
            n_threads (int): the number of threads to use.

        Returns:
            loglik (np.ndarray): A float64 array of shape (x.shape[0])
                where each element is the log-likelihood of that
                datapoint given the model.

        Raises:
            ValueError: Raised if unexpected inputs are supplied.
        """
        self._check_input_array(xdata, n_threads)
        if self.mu_mix is None or self.mix_weights is None:
            raise ValueError("Model not fitted yet.")

        resp = np.zeros((self.log_mu_mix.shape[0], xdata.shape[0]))
        if mask is not None or mask_terminal_dels or mask_gaps:
            xmasked = self.prep_masked_array(xdata, mask, mask_terminal_dels, mask_gaps)
            getProbsCExt_masked(xmasked, self.log_mu_mix, resp, n_threads)
            resp += self.log_mix_weights[:,None]
            resp = logsumexp(resp, axis=0)
            if normalize_scores:
                resp /= (xmasked < 21).sum(axis=1)

        else:
            getProbsCExt(xdata, self.log_mu_mix, resp, n_threads)
            resp += self.log_mix_weights[:,None]
            resp = logsumexp(resp, axis=0)
            if normalize_scores:
                resp /= xdata.shape[1]

        return resp



#Used for determining the version if running setup.
#Only change if building a new version.
__version__ = "0.3"

from .numbering_tools import SingleChainAnnotator, PairedChainAnnotator, PyCTermFinder
#from .numbering_tools.consensus_update_tools import build_consensus_files
from .vj_tools.vj_gene_assignment import VJGeneTool
from .scoring_tools import SequenceScoringTool, HumanizationTool, LiabilitySearchTool
from .antpack_cli import run_cli_interface



"""Contains the CTermFinder class."""
import os
import gzip
from antpack.antpack_cpp_ext import PrefilteringTool



class PyCTermFinder(PrefilteringTool):

    def __init__(self):
        """Class constructor.
        """
        consensus_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                "consensus_data")
        kmer_dict = _load_nterm_kmers()
        super().__init__(consensus_path, kmer_dict)



def _load_nterm_kmers():
    """A convenience function for loading a dictionary of n-terminal
    kmers."""
    filepath = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                "consensus_data", "START_REGION_KMER_FILES.txt.gz")

    kmer_dict = {}
    kmers_to_ints = {"H":0, "K":1, "L":2}

    with gzip.open(filepath, "rt") as fhandle:
        for line in fhandle:
            chain, kmer = line.strip().split(",")
            kmer_dict[kmer] = kmers_to_ints[chain]
    return kmer_dict



"""Contains default params shared by all schemes. Currently this comprises
only gap penalties for the N-terminal. This region is tricky because there
are frequenly large N-terminal deletions -- and also sometimes 
deletions located 10 AAs or so in from the N-terminal. We use
'staircase' gap penalties to encourage the aligner to handle this
sensibly. In any case where positions present in this gap penalty list
are also present in a scheme-specific constants file, the scheme-specific
file takes precedence."""

n_terminal_gap_positions = {}

for i in range(1,10):
    n_terminal_gap_positions[i] = [-1 - i, -25]

for i in range(10,18):
    n_terminal_gap_positions[i] = [-11 - (i-10) * 0.1, -25]


############################The following are alignment parameters that
#were selected to try to ensure a good alignment. Making large changes
#to these WILL change the results, possibly substantially.

DEFAULT_QUERY_GAP_PENALTY = -25
DEFAULT_TEMPLATE_GAP_PENALTY = -25

#The penalty for inserting a gap at a highly conserved position.
HIGHLY_CONSERVED_GAP_PENALTY = -65
#The bonus for complying at a highly conserved position.
HIGHLY_CONSERVED_BONUS = 65



"""Tests basic functionality for the SingleChainAnnotator class."""
import os
import re
import random
import gzip
import copy
import unittest
from antpack import SingleChainAnnotator

class TestSingleChainAnnotator(unittest.TestCase):


    def test_error_checking(self):
        """Check that sequences which have known issues are flagged
        as such, and that deliberately invalid inputs are recognized."""
        # Pass dummy sequences with errors.
        aligner = SingleChainAnnotator(chains=["H", "K", "L"])
        results = aligner.analyze_seqs(["YaY"])
        self.assertTrue(results[0][3].startswith("Sequence contains invalid"))
        results = aligner.analyze_seqs(["YBW"])
        self.assertTrue(results[0][3].startswith("Sequence contains invalid"))
        results = aligner.analyze_seqs(["Y K"])
        self.assertTrue(results[0][3].startswith("Sequence contains invalid"))
        results = aligner.analyze_seqs(["Y-K"])
        self.assertTrue(results[0][3].startswith("Sequence contains invalid"))

        results = aligner.analyze_seq("Y-K")
        self.assertTrue(results[3].startswith("Sequence contains invalid"))
        results = aligner.analyze_seq("yAy")
        self.assertTrue(results[3].startswith("Sequence contains invalid"))

        with self.assertRaises(RuntimeError):
            sorted_positions = aligner.sort_position_codes(["a", "1"])
        with self.assertRaises(RuntimeError):
            sorted_positions = aligner.sort_position_codes(["1", "2", "C3"])


    def test_chain_recognition(self):
        """Ensure that single chain annotator can correctly recognize the
        input chain when supplied with something that could be L or H,
        and ensure it can correctly detect sequences with large deletions
        that remove one or more conserved residues."""
        known_K = ("DIVMTQSPSSLTVTAGEKVTMSCKSSQSLLSSGNQKNYLTWYQQIPGQPPKLLIYWASTR"
                    "ESGVPDRFTGSGSGTDFTLTINSVQAEDLAVYYCQNDYTYPLTFGAGTKLELKRTV")
        known_L = ("QSALTQPASVSGSPGQSITISCTGTTSDVGTYNFVSWYQQHPGKAPKAIIFDVTNRPSGI"
                    "SNRFSGSKFGNTASLTISGLQAEDEADYYCAAYTVASTLLFGGGTKVTVLRQP")
        known_H = ("VHLQQSGAELMKPGASVKISCKASGYTFITYWIEWVKQRPGHGLEWIGDILPGSGSTNYN"
                    "ENFKGKATFTADSSSNTAYMQLSSLTSEDSAVYYCARSGYYGNSGFAYWGQGTLVTVSA")

        for multithreading_setting in [False]:
            for scheme in ["martin", "imgt", "kabat", "aho"]:
                aligner = SingleChainAnnotator(chains=["H", "K", "L"],
                            scheme = scheme)
                results = aligner.analyze_seqs([known_K, known_L, known_H])
                self.assertTrue(results[0][2] == "K")
                self.assertTrue(results[1][2] == "L")
                self.assertTrue(results[2][2] == "H")

                self.assertTrue(aligner.analyze_seq(known_K)[2] == "K")
                self.assertTrue(aligner.analyze_seq(known_L)[2] == "L")
                self.assertTrue(aligner.analyze_seq(known_H)[2] == "H")

                bad_chain = known_H[:100]
                self.assertTrue(aligner.analyze_seqs([bad_chain])[0][3].startswith("Unexpected"))
                self.assertTrue(aligner.analyze_seq(bad_chain)[3].startswith("Unexpected"))

    def test_performance(self):
        """Run a batch of test data (approximately 1600 sequences from the
        PDB) to ensure that numbering is consistent with numbering generated
        by another tool. There will occasionally be small differences in
        cases where there are multiple possible acceptable alignments,
        but in general we expect the numbering to be the same the vast
        majority of the time. Also check that the sequences can be correctly
        formed into an MSA."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs, martin_num, imgt_num, kabat_num, aho_num = \
                    [], [], [], [], []
            for line in fhandle:
                line_elements = line.strip().split(",")
                seqs.append(line_elements[0])
                martin_num.append(line_elements[1].split("_"))
                imgt_num.append(line_elements[2].split("_"))
                kabat_num.append(line_elements[3].split("_"))
                aho_num.append(line_elements[4].split("_"))

        os.chdir(current_dir)

        numberings = [martin_num, kabat_num, imgt_num, aho_num]
        schemes = ["martin", "kabat", "imgt", "aho"]

        for multithreading_setting in [False]:
            for scheme, numbering in zip(schemes, numberings):
                aligner = SingleChainAnnotator(chains=["H", "K", "L"],
                    scheme=scheme)
                aligner_results = aligner.analyze_seqs(seqs)
                total_comparisons, num_correct = compare_results(aligner_results,
                                    numbering, seqs, scheme)
                print(f"{scheme}: Total comparisons: {total_comparisons}. "
                        f"Num matching: {num_correct}. Multithreading {multithreading_setting}.")
                self.assertTrue(num_correct / total_comparisons > 0.97)

        # The last one produced is IMGT. Use this to test MSA construction.
        hnumbering, lnumbering, hseqs, lseqs = [], [], [], []

        for seq, numbering in zip(seqs, aligner_results):
            if numbering[2] == "H":
                hnumbering.append(numbering)
                hseqs.append(seq)
            elif numbering[2] in ["K", "L"]:
                lnumbering.append(numbering)
                lseqs.append(seq)

        observed_positions = set()
        for n in hnumbering:
            for pos in n[0]:
                observed_positions.add(pos)

        hpositions, hmsa = aligner.build_msa(hseqs, hnumbering, False)
        lpositions, lmsa = aligner.build_msa(lseqs, lnumbering, False)

        for position_set, msa in [(hpositions, hmsa), (lpositions, lmsa)]:
            for msa_seq in msa:
                self.assertTrue(len(msa_seq) == len(position_set))




    def test_alignment_trimming(self):
        """Make sure the alignment trimming procedure yields correct results."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]

        padded_seqs = []
        AAs = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"]

        random.seed(123)

        for seq in seqs:
            left_padding = random.randint(0,10)
            right_padding = random.randint(0,10)
            padded_seq = "".join([random.choice(AAs) for i in range(left_padding)]) + \
                    seq + "".join([random.choice(AAs) for i in range(right_padding)])
            padded_seqs.append(padded_seq)

        os.chdir(current_dir)
        aligner = SingleChainAnnotator(chains=["H", "K", "L"],
                    scheme="imgt")
        aligner_results = aligner.analyze_seqs(padded_seqs)
        
        for sequence, alignment in zip(padded_seqs, aligner_results):
            numbering = alignment[0]
            exstart = next((i for i in range(len(numbering)) if numbering[i] != '-'), 0)
            exend = next((i for i in range(exstart + 1, len(numbering)) if numbering[i] == '-'),
                    len(numbering))
            trimmed_sequence = sequence[exstart:exend]
            trimmed_numbering = numbering[exstart:exend]

            eval_seq, eval_numbering, eval_start, eval_end = aligner.trim_alignment(sequence,
                    alignment)

            self.assertTrue(eval_seq == trimmed_sequence)
            self.assertTrue(eval_seq == sequence[eval_start:eval_end])
            self.assertTrue(eval_numbering == trimmed_numbering)



    def test_region_labeling(self):
        """Ensure that the region labels assigned by the region labeling
        procedure correspond to our expectations, using a fairly
        inefficient procedure to determine ground-truth labeling."""
        regex = re.compile(r"^(?P<numbers>\d*)(?P<letters>\w*)$")

        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]

        imgt_labels = [(str(i), "fmwk1") for i in range(1,27)] + \
                [(str(i), "cdr1") for i in range(27,40)] + \
                [(str(i), "fmwk2") for i in range(39,56)] + \
                [(str(i), "cdr2") for i in range(56,66)] + \
                [(str(i), "fmwk3") for i in range(66,105)] + \
                [(str(i), "cdr3") for i in range(105,118)] + \
                [(str(i), "fmwk4") for i in range(118,129)]
        imgt_labels = {a:k for (a,k) in imgt_labels}

        martin_heavy = [(str(i), "fmwk1") for i in range(1,26)] + \
                [(str(i), "cdr1") for i in range(26,33)] + \
                [(str(i), "fmwk2") for i in range(33,52)] + \
                [(str(i), "cdr2") for i in range(52,57)] + \
                [(str(i), "fmwk3") for i in range(57,95)] + \
                [(str(i), "cdr3") for i in range(95,103)] + \
                [(str(i), "fmwk4") for i in range(103,114)]
        martin_heavy = {a:k for (a,k) in martin_heavy}
        martin_light = [(str(i), "fmwk1") for i in range(1,26)] + \
                [(str(i), "cdr1") for i in range(26,33)] + \
                [(str(i), "fmwk2") for i in range(33,50)] + \
                [(str(i), "cdr2") for i in range(50,53)] + \
                [(str(i), "fmwk3") for i in range(53,91)] + \
                [(str(i), "cdr3") for i in range(91,97)] + \
                [(str(i), "fmwk4") for i in range(97,108)]
        martin_light = {a:k for (a,k) in martin_light}

        kabat_heavy = [(str(i), "fmwk1") for i in range(1,31)] + \
                [(str(i), "cdr1") for i in range(31,36)] + \
                [(str(i), "fmwk2") for i in range(36,50)] + \
                [(str(i), "cdr2") for i in range(50,66)] + \
                [(str(i), "fmwk3") for i in range(66,95)] + \
                [(str(i), "cdr3") for i in range(95,103)] + \
                [(str(i), "fmwk4") for i in range(103,114)]
        kabat_heavy = {a:k for (a,k) in kabat_heavy}
        kabat_light = [(str(i), "fmwk1") for i in range(1,24)] + \
                [(str(i), "cdr1") for i in range(24,35)] + \
                [(str(i), "fmwk2") for i in range(35,50)] + \
                [(str(i), "cdr2") for i in range(50,57)] + \
                [(str(i), "fmwk3") for i in range(57,89)] + \
                [(str(i), "cdr3") for i in range(89,98)] + \
                [(str(i), "fmwk4") for i in range(98,108)]
        kabat_light = {a:k for (a,k) in kabat_light}



        scheme_labels = {"imgt":{"H":imgt_labels, "K":imgt_labels,
                "L":imgt_labels},
            "martin":{"H":martin_heavy, "K":martin_light,
                "L":martin_light},
            "kabat":{"H":kabat_heavy, "K":kabat_light,
                "L":kabat_light}
            }


        def get_gt_regions(numbering, label_map):
            gt_reg = []
            for n in numbering:
                if n == "-":
                    gt_reg.append("-")
                else:
                    gt_reg.append(label_map[regex.search(n).groups()[0]])
            return gt_reg

        os.chdir(current_dir)

        for scheme in ["imgt", "martin", "kabat"]:
            aligner = SingleChainAnnotator(chains=["H", "K", "L"],
                    scheme=scheme)
            num_err = 0

            for seq in seqs:
                numbering = aligner.analyze_seq(seq)
                labels = aligner.assign_cdr_labels(numbering)

                gt_regions = get_gt_regions(numbering[0],
                        scheme_labels[scheme][numbering[2]])
                if gt_regions != labels:
                    num_err += 1
            self.assertTrue(num_err == 0)


    def test_position_code_sorting(self):
        """Checks the position code sorting function to make
        sure it is sorting positions correctly for different schemes."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]

        os.chdir(current_dir)
        random.seed(123)

        num_err = 0

        for scheme in ["martin", "imgt", "kabat", "aho"]:
            aligner = SingleChainAnnotator(chains=["H", "K", "L"], scheme=scheme)
            for seq in seqs:
                numbering = aligner.analyze_seq(seq)[0]
                numbering = [n for n in numbering if n != "-"]
                shuffled_numbering = copy.deepcopy(numbering)
                random.shuffle(shuffled_numbering)
                sorted_numbering = aligner.sort_position_codes(shuffled_numbering)
                if numbering != sorted_numbering:
                    num_err += 1

        self.assertTrue(num_err == 0)



    def test_extended_length_handling(self):
        """Check situations where the sequence is much longer than
        a typical variable chain."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        AAs = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"]

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]

        os.chdir(current_dir)
        random.seed(123)

        aligner = SingleChainAnnotator(chains=["H", "K", "L"], scheme="imgt")
        for seq in seqs:
            alignment = aligner.analyze_seq(seq)
            # Exclude sequences that have gaps at the ends
            if alignment[-1].startswith('Unexpected AA'):
                continue
            if '120' not in alignment[0]:
                continue
            if alignment[2] == "H":
                if '128' not in alignment[0]:
                    continue
                if seq[alignment[0].index('128')] == "G":
                    continue
            else:
                if '127' not in alignment[0]:
                    continue

            muddled_seq = seq + "YYYGGGGGSSSS" + "".join([random.choice(AAs) for i in range(250)])
            alt_alignment = aligner.analyze_seq(muddled_seq)
            self.assertTrue(len(alt_alignment[0]) == len(muddled_seq))


            trimmed_alt = aligner.trim_alignment(muddled_seq, alt_alignment)
            trimmed_gt = aligner.trim_alignment(seq, alignment)
            self.assertTrue(trimmed_alt[0] == trimmed_gt[0])



    def test_x_handling(self):
        """Check situations where one or more letters has
        been replaced with X."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        AAs = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"]

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]

        os.chdir(current_dir)
        random.seed(123)

        aligner = SingleChainAnnotator(chains=["H", "K", "L"], scheme="imgt")
        for seq in seqs:
            pre_x_alignment = aligner.analyze_seq(seq)
            if pre_x_alignment[-1] != "":
                continue
            if '1' not in pre_x_alignment[0] or '73' not in pre_x_alignment:
                continue

            xposition = random.randint(0, len(seq) - 1)
            seq_list = list(seq)
            seq_list[xposition] = "X"
            alignment = aligner.analyze_seq("".join(seq_list))

            self.assertTrue(alignment[0] == pre_x_alignment[0])



def compare_results(results, comparator_numbering, seqs, scheme):
    """Compares the numbering generated by AntPack with a comparator,
    and returns the number correct vs total comparisons. Also contains
    some (commented-out) code for writing non-matching results to
    a temporary file for closer inspection."""
    total_comparisons, num_correct = 0, 0
    #outhandle = open(f"temp_{scheme}.csv", "w+", encoding="utf-8")
    for result, comparator, seq in zip(results, comparator_numbering, seqs):
        if result[3] != '':
            continue
        total_comparisons += 1
        if result[0] == comparator:
            num_correct += 1
        #This code writes results to file in a format which is easy to look
        #at. Comment it out normally.
        #else:
        #    for i, resnum in enumerate(result[0]):
        #        if resnum != comparator[i]:
        #            result[0][i] = result[0][i] + "!!"
        #    outhandle.write(f"Sequence,{','.join(list(seq))}\n")
        #    outhandle.write(f"Antpack_result,{','.join(result[0])}\n")
        #    outhandle.write(f"Comparator_result,{','.join(comparator)}\n\n")

    #outhandle.close()
    return total_comparisons, num_correct


if __name__ == "__main__":
    unittest.main()



"""Simple tools for reading biological data files. This avoids
any need to use Biopython or other similar libraries as dependencies."""
import gzip


def read_fasta(filepath):
    """Generator for reading fasta files. By default, reads
    one sequence at a time. Can take either gzipped or non-
    gzipped files as input. Assumes utf-8 encoding."""
    if filepath.endswith(".gz"):
        fhandle = gzip.open(filepath, "rt", encoding="utf-8")
    else:
        fhandle = open(filepath, "r", encoding="utf-8")

    sequence, seqinfo = [], ""

    for line in fhandle:
        if line.startswith(">"):
            if len(sequence) > 0 and len(seqinfo) > 0:
                yield seqinfo, "".join(sequence)
                sequence, seqinfo = [], line[1:].strip()

            elif len(sequence) == 0 and len(seqinfo) == 0:
                seqinfo = line[1:].strip()

            else:
                raise RuntimeError("Incorrect fasta file formatting or missing "
                        f"sequence id / name; please check input file {filepath}.")

        elif len(line.strip()) == 0:
            continue

        else:
            sequence.append(line.strip())

    if len(sequence) > 0 and len(seqinfo) > 0:
        yield seqinfo, "".join(sequence)
    elif len(sequence) != 0 or len(seqinfo) != 0:
        raise RuntimeError("Incorrect fasta file formatting or missing "
                f"sequence id / name; please check input file {filepath}.")

    fhandle.close()



"""Tests basic functionality for the PairedChainAnnotator class.
Notice that this test relies on the correct functioning of
SingleChainAnnotator, so if test_single_chain_annotator does
not pass, this one will not either."""
import os
import random
import gzip
import unittest
from antpack import PairedChainAnnotator, SingleChainAnnotator
from antpack.scoring_tools.scoring_constants import scoring_constants as SCCONST



class TestPairedChainAnnotator(unittest.TestCase):


    def test_error_checking(self):
        """Check that sequences which have known issues are flagged
        as such, and that deliberately invalid inputs are recognized."""
        # Pass dummy sequences with errors.
        aligner = PairedChainAnnotator()

        results = aligner.analyze_seq("YaY")
        self.assertTrue(results[0][3].startswith("Invalid sequence")
                or results[0][3].startswith("Sequence contains"))
        self.assertTrue(results[1][3].startswith("Invalid sequence")
                or results[1][3].startswith("Sequence contains"))

        results = aligner.analyze_seq("YBW")
        self.assertTrue(results[0][3].startswith("Invalid sequence")
                or results[0][3].startswith("Sequence contains"))
        self.assertTrue(results[1][3].startswith("Invalid sequence")
                or results[1][3].startswith("Sequence contains"))

        results = aligner.analyze_seq("Y K")
        self.assertTrue(results[0][3].startswith("Invalid sequence")
                or results[0][3].startswith("Sequence contains"))
        self.assertTrue(results[1][3].startswith("Invalid sequence")
                or results[1][3].startswith("Sequence contains"))

        results = aligner.analyze_seq("Y-K")
        self.assertTrue(results[0][3].startswith("Invalid sequence")
                or results[0][3].startswith("Sequence contains"))
        self.assertTrue(results[1][3].startswith("Invalid sequence")
                or results[1][3].startswith("Sequence contains"))



    def test_analyze_seqs(self):
        """Test that analyze_seqs returns the same
        results as analyze_seq applied to the list.
        (analyze_seqs calls analyze_seq, so highly
        unlikely it does anything different...but just
        in case)."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))
        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]
        os.chdir(current_dir)

        sc_aligner = SingleChainAnnotator(scheme="imgt")
        pc_aligner = PairedChainAnnotator(scheme="imgt")

        alignments = [sc_aligner.analyze_seq(seq) for seq in seqs]

        heavy_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] == "H"]
        light_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] != "H"]

        # Join the sequences with a very arbitrary linker.
        merged_seqs = [l[0] + "YYYSSSGGG" + h[0] for
                h, l in zip(heavy_chains, light_chains)]

        heavy_test, light_test = pc_aligner.analyze_seqs(merged_seqs)
        self.assertTrue(len(heavy_test)==len(merged_seqs))
        self.assertTrue(len(light_test)==len(merged_seqs))

        for i, seq in enumerate(merged_seqs):
            h, l = pc_aligner.analyze_seq(seq)
            self.assertTrue(h==heavy_test[i])
            self.assertTrue(l==light_test[i])




    def test_sequence_extraction(self):
        """Take test heavy and light chains and number them using
        SingleChainAnnotator. Next, join pairs of chains with a random
        number of random amino acids and add a random number of
        random amino acids at the beginning and end. Compare the
        numbering assigned by SingleChainAnnotator to that assigned
        by PairedChainAnnotator (this of course assumes that
        SingleChainAnnotator is functioning correctly)."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))
        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]
        os.chdir(current_dir)

        sc_aligner = SingleChainAnnotator(scheme="imgt")
        m_aligner = PairedChainAnnotator(scheme="imgt")

        alignments = [sc_aligner.analyze_seq(seq) for seq in seqs]

        heavy_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] == "H"]
        light_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] != "H"]

        random.seed(0)

        # Try with both light chain first and heavy chain first.
        order1, order2 = list(zip(heavy_chains, light_chains)), \
                list(zip(light_chains, heavy_chains))

        for ordering in (order1, order2):
            for hc, lc in ordering:
                prefix = [SCCONST.aa_list[random.randint(0,19)]
                    for i in range(random.randint(5,25))]
                suffix = [SCCONST.aa_list[random.randint(0,19)]
                    for i in range(random.randint(5,25))]
                joiner1 = [SCCONST.aa_list[random.randint(0,19)]
                    for i in range(random.randint(5,25))]
                joiner2 = [SCCONST.aa_list[random.randint(0,19)]
                    for i in range(random.randint(5,25))]

                merged_hc = "".join( ["".join(prefix), hc[0], "".join(joiner1) ] )
                merged_lc = "".join( ["".join(joiner2), lc[0], "".join(suffix) ] )
                merged_chain = merged_hc + merged_lc

                hc_align = sc_aligner.analyze_seq(merged_hc)
                lc_align = sc_aligner.analyze_seq(merged_lc)
                if hc_align[1] < 0.7 or lc_align[1] < 0.7:
                    continue

                if hc_align[-1] != "" or lc_align[-1] != "":
                    continue

                for i, lcpos in enumerate(lc_align[0]):
                    if lcpos != "-":
                        break
                for j, lcpos in reversed(list(enumerate(lc_align[0]))):
                    if lcpos != "-":
                        break

                trimmed_lc_seq = merged_lc[i:j+1]
                trimmed_lc_align = lc_align[0][i:j+1]
                if not trimmed_lc_align[0] == "1":
                    continue

                for i, hcpos in enumerate(hc_align[0]):
                    if hcpos != "-":
                        break
                for j, hcpos in reversed(list(enumerate(hc_align[0]))):
                    if hcpos != "-":
                        break

                trimmed_hc_seq = merged_hc[i:j+1]
                trimmed_hc_align = hc_align[0][i:j+1]
                if not trimmed_hc_align[-1] == "128":
                    continue

                mc_heavy, mc_light = m_aligner.analyze_seq(merged_chain)
                self.assertTrue(len(mc_heavy[0]) == len(merged_chain))
                self.assertTrue(len(mc_light[0]) == len(merged_chain))

                _, mchn, hstart, hend = m_aligner.trim_alignment(merged_chain, mc_heavy)
                _, mcln, lstart, lend = m_aligner.trim_alignment(merged_chain, mc_light)

                self.assertTrue(mcln == trimmed_lc_align)
                self.assertTrue(mchn == trimmed_hc_align)
                self.assertTrue(merged_chain[lstart:lend] ==
                        trimmed_lc_seq)
                self.assertTrue(merged_chain[hstart:hend] ==
                        trimmed_hc_seq)


    def test_single_chain_behavior(self):
        """Take known heavy / light chains only and run them 
        through PairedChainAnnotator. Compare the
        numbering assigned by SingleChainAnnotator to that assigned
        by PairedChainAnnotator (this of course assumes that
        SingleChainAnnotator is functioning correctly)."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))
        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line in fhandle]
        os.chdir(current_dir)

        sc_aligner = SingleChainAnnotator(scheme="imgt")
        m_aligner = PairedChainAnnotator(scheme="imgt")

        # Skip sequences with only one cysteine, which are hard
        # to analyze.
        seqs = [s for s in seqs if len([l for l in s if l == "C"])
                >= 2]

        alignments = [sc_aligner.analyze_seq(seq) for seq in seqs]

        heavy_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] == "H"]
        light_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] != "H"]

        random.seed(0)

        for (seq, alignment) in heavy_chains:
            pc_result = m_aligner.analyze_seq(seq)[0]
            self.assertTrue(pc_result == alignment)

        for (seq, alignment) in light_chains:
            pc_result = m_aligner.analyze_seq(seq)[1]
            self.assertTrue(pc_result == alignment)




if __name__ == "__main__":
    unittest.main()



"""Contains the SingleChainAnnotator class, which provides the tools needed
to parse single chains (either heavy or light) or a sequence which likely contains
a heavy or light chain."""
import os
from .cterm_finder import _load_nterm_kmers
from antpack.antpack_cpp_ext import SingleChainAnnotatorCpp




class SingleChainAnnotator(SingleChainAnnotatorCpp):

    def __init__(self, chains=["H", "K", "L"], scheme="imgt"):
        """Class constructor.

        Args:
            chains (list): A list of chains. Each must be one of "H", "K", "L".
                If ["H", "K", "L"] (default), the annotator will automatically
                determine the most appropriate chain type for each input
                sequence.
            scheme (str): The numbering scheme. Must be one of "imgt",
                "martin", "kabat", "aho".

        Raises:
            ValueError: A ValueError is raised if unacceptable inputs are
                supplied.
        """
        consensus_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                "consensus_data")
        kmer_dict = _load_nterm_kmers()
        super().__init__(chains, scheme, consensus_path, kmer_dict)



"""Contains the PairedChainAnnotator class, which extracts variable region
heavy and light chains from an input sequence that is presumed to contain
a single variable heavy and a single variable light chain (but may
contain only a single chain instead)."""
import os
from .cterm_finder import _load_nterm_kmers
from antpack.antpack_cpp_ext import PairedChainAnnotatorCpp



class PairedChainAnnotator(PairedChainAnnotatorCpp):

    def __init__(self, scheme = "imgt"):
        """Class constructor.

        Args:
            scheme (str): The numbering scheme. Must be one of "imgt",
                "martin", "kabat", "aho".

        Raises:
            RuntimeError: A RuntimeError is raised if unacceptable inputs are
                supplied.
        """
        consensus_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                "consensus_data")
        kmer_dict = _load_nterm_kmers()
        super().__init__(scheme, consensus_path, kmer_dict)



"""Contains special functions implemented here to avoid having to
introduce additional dependencies (logsumexp from scipy)."""
import numpy as np


def logsumexp(a, axis):
    """Compute the log of the sum of exponentials of input elements
    in a numerically stable way. This is a simplified version of
    scipy's logsumexp function.

    Args:
        a (np.ndarray): The input array.
        axis (int): The axis over which to apply this operation.

    Returns:
        out (np.ndarray): The result of the logsumexp operation.
    """
    initial_value = -np.inf if np.size(a) == 0 else None
    a_max = np.amax(a, axis=axis, keepdims=True, initial=initial_value)

    if a_max.ndim > 0:
        a_max[~np.isfinite(a_max)] = 0
    elif not np.isfinite(a_max):
        a_max = 0

    tmp = np.exp(a - a_max)

    # suppress warnings about log of zero
    with np.errstate(divide='ignore'):
        s = np.sum(tmp, axis=axis, keepdims=False)
        out = np.log(s)

    a_max = np.squeeze(a_max, axis=axis)
    out += a_max

    return out



"""Contains constants used across multiple procedures."""

#The list of aas; the position in the list determines the
#number assigned for one-hot encoding.
aa_list = ["A", "C", "D", "E", "F", "G", "H", "I", "K",
        "L", "M", "N", "P", "Q", "R", "S", "T", "V",
        "W", "Y", "-"]

#The allowed amino acids.
allowed_aas = set(aa_list)

#The minimum percent identity allowed. This is pretty generous...most should
#be much higher.
MIN_PERCENT_IDENTITY = 0.8

#The number of positions required to be non-blank. Fairly generous.
#MIN_HEAVY_NOT_BLANK = 110
MIN_HEAVY_NOT_BLANK = 105
MIN_LIGHT_NOT_BLANK = 105


#The rescaling parameters for heavy and light chain scores
#when combining them. Based on training set.
HEAVY_MEDIAN_SCORE = -59.5617
LIGHT_MEDIAN_SCORE = -33.22445



"""Tests scoring outputs of the c extension by comparing them
to a slow but easy to cross-check Python routine."""
import os
import unittest
import numpy as np
import pandas as pd
from antpack.utilities.special_functions import logsumexp
from antpack import SequenceScoringTool
from antpack.antpack_cpp_ext import mask_terminal_deletions


class TestCExtScoring(unittest.TestCase):


    def test_cext_scoring(self):
        """Check the workhorse function that provides scores
        for consistency with a simple Python version."""
        start_dir = os.path.abspath(os.path.dirname(__file__))
        score_tool = SequenceScoringTool()
        raw_data = pd.read_csv(os.path.join(start_dir, "test_data",
            "imgt_comp_scoring.csv.gz"))

        heavy_chains = raw_data[raw_data["chain_types"]=="H"]
        light_chains = raw_data[raw_data["chain_types"].isin(["K", "L"])]

        heavy_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in heavy_chains["sequences"].tolist()[:10]]
        light_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in light_chains["sequences"].tolist()[:10]]

        heavy_arr = np.vstack(heavy_arr)
        light_arr = np.vstack(light_arr)
        heavy_model = score_tool.models["human"]["H"]
        light_model = score_tool.models["human"]["L"]

        raw_heavy_score = heavy_model.score(heavy_arr, n_threads = 2)
        raw_light_score = light_model.score(light_arr, n_threads = 2)

        comp_heavy_score = np.zeros((heavy_model.n_components, raw_heavy_score.shape[0]))
        comp_light_score = np.zeros((light_model.n_components, raw_light_score.shape[0]))

        for i in range(heavy_arr.shape[0]):
            for k in range(heavy_model.log_mu_mix.shape[0]):
                for j in range(heavy_model.log_mu_mix.shape[1]):
                    comp_heavy_score[k,i] += heavy_model.log_mu_mix[k,j,heavy_arr[i,j]]

        comp_heavy_score += heavy_model.log_mix_weights[:,None]
        comp_heavy_score = logsumexp(comp_heavy_score, axis=0)
        self.assertTrue(np.allclose(comp_heavy_score, raw_heavy_score))

        for i in range(light_arr.shape[0]):
            for k in range(light_model.log_mu_mix.shape[0]):
                for j in range(light_model.log_mu_mix.shape[1]):
                    comp_light_score[k,i] += light_model.log_mu_mix[k,j,light_arr[i,j]]

        comp_light_score += light_model.log_mix_weights[:,None]
        comp_light_score = logsumexp(comp_light_score, axis=0)
        self.assertTrue(np.allclose(comp_light_score, raw_light_score))


    def test_terminal_deletions(self):
        """Tests the categorical mixture masked scoring function by
        comparing it with a simple python implementation."""
        start_dir = os.path.abspath(os.path.dirname(__file__))
        score_tool = SequenceScoringTool()
        raw_data = pd.read_csv(os.path.join(start_dir, "test_data",
            "imgt_comp_scoring.csv.gz"))

        heavy_chains = raw_data[raw_data["chain_types"]=="H"].sequences.tolist()
        light_chains = raw_data[raw_data["chain_types"].isin(["K", "L"])].sequences.tolist()

        heavy_chains[0] = heavy_chains[0][10:]
        heavy_chains[1] = heavy_chains[1][:-10]
        light_chains[0] = light_chains[0][10:]
        light_chains[1] = light_chains[1][:-25]

        heavy_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in heavy_chains[:10]]
        light_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in light_chains[:10]]
        heavy_arr = np.vstack(heavy_arr)
        light_arr = np.vstack(light_arr)

        dummy_heavy_arr = heavy_arr.copy()
        dummy_light_arr = light_arr.copy()
        mask_terminal_deletions(dummy_heavy_arr)
        mask_terminal_deletions(dummy_light_arr)

        for i in range(light_arr.shape[0]):
            for k in range(light_arr.shape[1]):
                if light_arr[i,k] != 20:
                    break
                light_arr[i,k] = 21
            for k in range(light_arr.shape[1]-1, 0, -1):
                if light_arr[i,k] != 20:
                    break
                light_arr[i,k] = 21

        for i in range(heavy_arr.shape[0]):
            for k in range(heavy_arr.shape[1]):
                if heavy_arr[i,k] != 20:
                    break
                heavy_arr[i,k] = 21
            for k in range(heavy_arr.shape[1]-1, 0, -1):
                if heavy_arr[i,k] != 20:
                    break
                heavy_arr[i,k] = 21

        self.assertTrue(np.allclose(dummy_heavy_arr, heavy_arr))
        self.assertTrue(np.allclose(dummy_light_arr, light_arr))



    def test_mask_scoring(self):
        """Ensure that the terminal deletion masking function yields
        expected results."""
        start_dir = os.path.abspath(os.path.dirname(__file__))
        score_tool = SequenceScoringTool()
        raw_data = pd.read_csv(os.path.join(start_dir, "test_data",
            "imgt_comp_scoring.csv.gz"))

        heavy_chains = raw_data[raw_data["chain_types"]=="H"]
        light_chains = raw_data[raw_data["chain_types"].isin(["K", "L"])]

        heavy_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in heavy_chains["sequences"].tolist()[:10]]
        light_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in light_chains["sequences"].tolist()[:10]]
        heavy_arr = np.vstack(heavy_arr)
        light_arr = np.vstack(light_arr)

        heavy_masks, light_masks = [], []

        heavy_masks = [(i < 10 or i > 20) for i in range(heavy_arr.shape[1])]
        light_masks = [(i < 10 or i > 20) for i in range(light_arr.shape[1])]

        heavy_model = score_tool.models["human"]["H"]
        light_model = score_tool.models["human"]["L"]

        comp_heavy_score = np.zeros((heavy_model.n_components, heavy_arr.shape[0]))
        comp_light_score = np.zeros((light_model.n_components, light_arr.shape[0]))

        for i in range(heavy_arr.shape[0]):
            for k in range(heavy_model.log_mu_mix.shape[0]):
                for j in range(heavy_model.log_mu_mix.shape[1]):
                    if not heavy_masks[j]:
                        continue
                    comp_heavy_score[k,i] += heavy_model.log_mu_mix[k,j,heavy_arr[i,j]]

        for i in range(light_arr.shape[0]):
            for k in range(light_model.log_mu_mix.shape[0]):
                for j in range(light_model.log_mu_mix.shape[1]):
                    if not light_masks[j]:
                        continue
                    comp_light_score[k,i] += light_model.log_mu_mix[k,j,light_arr[i,j]]

        heavy_masks = np.array(heavy_masks, dtype=bool)
        light_masks = np.array(light_masks, dtype=bool)

        raw_heavy_score = heavy_model.score(heavy_arr, heavy_masks, n_threads = 2)
        raw_light_score = light_model.score(light_arr, light_masks, n_threads = 2)

        comp_heavy_score += heavy_model.log_mix_weights[:,None]
        comp_heavy_score = logsumexp(comp_heavy_score, axis=0)
        self.assertTrue(np.allclose(comp_heavy_score, raw_heavy_score))

        comp_light_score += light_model.log_mix_weights[:,None]
        comp_light_score = logsumexp(comp_light_score, axis=0)
        self.assertTrue(np.allclose(comp_light_score, raw_light_score))



if __name__ == "__main__":
    unittest.main()



"""Tests basic functionality for the SequenceScoringTool class.
Assumes that pandas is installed."""
import os
import unittest
import numpy as np
import pandas as pd
from antpack import SequenceScoringTool, SingleChainAnnotator, HumanizationTool

class TestSequenceScoringTool(unittest.TestCase):


    def test_error_checking(self):
        """Check that invalid data passed to the sequence scoring
        tool raises expected exceptions."""
        score_tool = SequenceScoringTool()
        humanization_tool = HumanizationTool()

        # Sequences containing unrecognized characters should raise
        # a value error for these two routines, and return np.nan
        # for the other two.
        with self.assertRaises(ValueError):
            humanization_tool.suggest_mutations("WoW")

        self.assertTrue(score_tool.get_diagnostic_info("WoW")[2]=="unknown")
        self.assertTrue(np.isnan(score_tool.score_seqs(["WoW"])[0]))

        with self.assertRaises(ValueError):
            score_tool.models["human"]["H"].score(np.zeros((1,173)))
        with self.assertRaises(ValueError):
            score_tool.models["human"]["H"].score(np.zeros((173), dtype=np.uint8))
        with self.assertRaises(ValueError):
            score_tool.models["human"]["H"].score(np.zeros((1,141), dtype=np.uint8))

        with self.assertRaises(ValueError):
            score_tool.models["human"]["L"].score(np.zeros((1,141)))
        with self.assertRaises(ValueError):
            score_tool.models["human"]["L"].score(np.zeros((141), dtype=np.uint8))
        with self.assertRaises(ValueError):
            score_tool.models["human"]["L"].score(np.zeros((1,173), dtype=np.uint8))

        proba_test = score_tool.models["human"]["H"].predict(np.zeros((1,173), dtype=np.uint8),
                return_raw_probs = True)
        self.assertTrue(len(proba_test.shape) == 2)
        self.assertTrue(proba_test.shape[1] == 1)
        self.assertTrue(proba_test.shape[0] == score_tool.models["human"]["H"].n_components)


    def test_sequence_extraction(self):
        """Make sure the sequence prep feature is providing expected
        results."""
        score_tool = SequenceScoringTool()

        start_dir = os.path.abspath(os.path.dirname(__file__))

        raw_data = pd.read_csv(os.path.join(start_dir, "test_data",
            "imgt_comp_scoring.csv.gz"))

        aligners = {"H":SingleChainAnnotator(["H"]),
            "K":SingleChainAnnotator(["K"]),
            "L":SingleChainAnnotator(["L"])  }

        for seq, chain_name in zip(raw_data["sequences"].tolist()[:10],
                raw_data["chain_types"].tolist()[:10]):
            p_extract, assigned_chain, _, _, _, _ = score_tool._full_prep_sequence(seq)
            if chain_name == "K":
                self.assertTrue(assigned_chain == "L")
            else:
                self.assertTrue(assigned_chain == chain_name)

            numbering = aligners[assigned_chain].analyze_seq(seq)[0]
            position_dict = score_tool.position_dict[assigned_chain]
            seq_extract = ["-" for i in range(len(position_dict))]

            for position, aa in zip(numbering, seq):
                if position == "-" or position not in position_dict:
                    continue
                seq_extract[position_dict[position]] = aa

            self.assertTrue(seq_extract == p_extract)




    def test_scoring_consistency(self):
        """Test that scores assigned by different procedures give what
        we expect."""
        score_tool = SequenceScoringTool(offer_classifier_option=True)
        adj_score_tool = SequenceScoringTool(offer_classifier_option=False,
                normalization="training_set_adjust")
        adj_score_tool.aligner = SingleChainAnnotator()

        start_dir = os.path.abspath(os.path.dirname(__file__))

        raw_data = pd.read_csv(os.path.join(start_dir, "test_data",
            "imgt_comp_scoring.csv.gz"))

        heavy_chains = raw_data[raw_data["chain_types"]=="H"]
        light_chains = raw_data[raw_data["chain_types"].isin(["K", "L"])]

        heavy_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in heavy_chains["sequences"].tolist()]
        light_arr = [score_tool._simple_prep_sequence(s)[1] for s
                in light_chains["sequences"].tolist()]

        heavy_arr = np.vstack(heavy_arr)
        light_arr = np.vstack(light_arr)

        # Generate scores using a variety of different procedures. We will compare
        # these to make sure everything makes sense.

        raw_heavy_score = score_tool.models["human"]["H"].score(heavy_arr, n_threads = 2)
        raw_light_score = score_tool.models["human"]["L"].score(light_arr, n_threads = 2)

        unadj_heavy_score = score_tool.score_seqs(heavy_chains["sequences"].tolist())
        unadj_light_score = score_tool.score_seqs(light_chains["sequences"].tolist())
        self.assertTrue(np.allclose(unadj_heavy_score, raw_heavy_score))
        self.assertTrue(np.allclose(unadj_light_score, raw_light_score))

        og_term_del_scores = raw_data["term_del_scores"].values
        term_del_scores = adj_score_tool.score_seqs(raw_data["sequences"].tolist(),
                mask_terminal_dels=True)
        self.assertTrue(np.allclose(term_del_scores, og_term_del_scores))



if __name__ == "__main__":
    unittest.main()



"""Defines important defaults for the Martin numbering scheme."""

#The expected length of the heavy and light chains
NUM_HEAVY = 113
NUM_LIGHT = 107

############################The following are alignment parameters that
#were selected to try to ensure a good alignment. Making large changes
#to these WILL change the results, possibly substantially.

#These are positions that are essentially mandatory. A sequence that deviates
#from one of these is an alignment issue or has a very large deletion.
heavy_conserved_positions = {22:"C", 36:"W", 92:"C", 103:"W", 104:"G", 106:"G"}
light_conserved_positions = {23:"C", 35:"W", 88:"C", 98:"F", 99:"G", 101:"G"}

#These are positions where either A) a blank in the query sequence is very common or
#B) insertions are very common. We want special template and query gap penalties for
#these positions. One or two of these are chain dependent. The first value is
#the query gap column, the second is the template gap column.
heavy_special_positions = {8:[-25,-1.0], 31:[-1.1,-1.1],
        40:[-11.4,-25], 41:[-11.3,-25], 42:[-11.2,-25], 43:[-11.1,-25], 44:[-11,-25],
        52:[-1.1,-1.1], 72:[-25,-1.0], 100:[-1.1,-1.1]}

light_special_positions = {10:[-1,-11.0], 68:[-1.1,-1.1],
        #CDR 1
        30:[-11.1,-1.0],
        #CDR 2, 3
        52:[-1.1,-1.1], 95:[-1.1,-1.1]}

# These exclude special positions. For Martin, unlike IMGT, heavy and light CDRs are different.
heavy_cdrs = {26:-11.5, 27:-11.4, 28:-11.3, 29:-11.2, 30:-11.1,
        # CDR 2
        50:-11.2, 51:-11.1, 53:-11.3, 54:-11.4, 55:-11.5, 56:-11.6,
        #CDR 3
        95:-11.5, 96:-11.4, 97:-11.3, 98:-11.2, 99:-11.1,
            101:-11.6, 102:-11.7}
light_cdrs = {26:-11.5, 27:-11.4, 28:-11.3, 29:-11.2, 31:-11.0,
        # CDR 2
        50:-11.2, 51:-11.1, 53:-11, 54:-11.15, -55:-11.25, 56:-11.35,
        #CDR 3
        89:-11.5, 90:-11.4, 91:-11.3, 92:-11.2, 93:-11.1, 94:-11, 96:-11.05, 97:-11.15}



import platform
import gc
import pytest

is_pypy = platform.python_implementation() == 'PyPy'
is_darwin = platform.system() == 'Darwin'

def collect() -> None:
    if is_pypy:
        for _ in range(3):
            gc.collect()
    else:
        gc.collect()

skip_on_pypy = pytest.mark.skipif(
    is_pypy, reason="This test currently fails/crashes PyPy")

xfail_on_pypy_darwin = pytest.mark.xfail(
    is_pypy and is_darwin, reason="This test for some reason fails on PyPy/Darwin")



import argparse
import sys
import sysconfig  # type: ignore

from . import __version__, include_dir, cmake_dir


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the version number.",
    )
    parser.add_argument(
        "--include_dir",
        action="store_true",
        help="Print the path to the nanobind C++ header directory."
    )
    parser.add_argument(
        "--cmake_dir",
        action="store_true",
        help="Print the path to the nanobind CMake module directory."
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.include_dir:
        print(include_dir())
    if args.cmake_dir:
        print(cmake_dir())


if __name__ == "__main__":
    main()



"""This module contains tools needed to assign the closest VJ germline
VJ gene (highest percent identity), using amino acid information
only. Bear in mind that tools which offer a probabilistic assigment
(e.g. IGOR) may be more informative."""
import os
import numpy as np
from ..utilities.vj_utilities import load_vj_gene_consensus_db
from antpack.antpack_cpp_ext import VJMatchCounter



class VJGeneTool(VJMatchCounter):
    """Contains functionality needed to find the closest
    VJ genes for a given amino acid sequence and retrieve the
    sequence for those VJ genes. The date of the database
    used for these assignments is stored and can be retrieved
    if needed."""

    def __init__(self, database = "imgt", scheme = "imgt"):
        """Class constructor.

        Args:
            database (str): One of 'imgt', 'ogrdb'. Determines whether
                the IMGT or OGRDB databases are used for germline gene
                assignment.
            scheme (str): One of 'aho, 'imgt', 'kabat' or 'martin'.
                Determines the numbering scheme that will be used
                when assigning vj genes based on alignments.
        """
        if database not in ("imgt", "ogrdb"):
            raise RuntimeError("Invalid database supplied.")
        project_path = os.path.abspath(os.path.dirname(__file__))
        db_path = os.path.join(project_path, "consensus_data")
        vj_names, vj_seqs, retrieved_dates = load_vj_gene_consensus_db(os.getcwd(),
                db_path, database)

        blosum_matrix = np.load(os.path.join(project_path, "..",
            "numbering_tools", "consensus_data",
            "blosum_matrix.npy")).astype(np.float64)
        self.retrieved_dates = retrieved_dates
        ig_aligner_consensus_path = os.path.join(project_path, "..", "numbering_tools",
                "consensus_data")
        super().__init__(vj_names, vj_seqs, blosum_matrix,
                scheme, ig_aligner_consensus_path)



    def retrieve_db_dates(self):
        """Returns the dates when each VJ gene database
        used for this assignment was last updated by downloading
        from IMGT or OGRDB."""
        return self.retrieved_dates



"""Contains utilities useful for loading fitted models."""
import os
import gzip
import numpy as np
from ..scoring_tools.categorical_mix import CategoricalMixture
from ..scoring_tools.scoring_constants import allowed_imgt_pos as ahip


def load_model(start_dir, chain_type="heavy", species="human"):
    """Loads the model for the specified species."""
    current_dir = os.getcwd()
    os.chdir(os.path.join(start_dir, "model_data"))

    if chain_type == "heavy":
        sequence_length = len(ahip.heavy_allowed_positions)
    elif chain_type == "light":
        sequence_length = len(ahip.light_allowed_positions)
    else:
        raise ValueError("Unrecognized chain type supplied.")

    try:
        mu = np.load(f"{species}_{chain_type}_mu.npy").astype(np.float64)
        mixweights = np.load(f"{species}_{chain_type}_mixweights.npy").astype(np.float64)
    except Exception as exc:
        raise ValueError("Final run not present in expected location.") from exc

    model = CategoricalMixture(mixweights.shape[0], num_possible_items = 21,
                         sequence_length = sequence_length)

    model.load_params(mu, mixweights)

    os.chdir(current_dir)
    return model



"""Contains utilities useful for compiling the raw sequence data
and encoding."""
from ..scoring_tools.scoring_constants import allowed_imgt_pos as ahip

def get_position_dict(chain_type="heavy"):
    """Loads the dictionary mapping IMGT numbers to positions."""
    if chain_type == "heavy":
        positions = ahip.heavy_allowed_positions
    elif chain_type == "light":
        positions = ahip.light_allowed_positions
    else:
        raise ValueError("Unrecognized chain type passed.")

    position_dict = {pos:i for (i, pos) in enumerate(positions)}
    return position_dict, len(position_dict)

def get_reverse_position_dict(chain_type="heavy"):
    """Sets up a dictionary mapping positions to IMGT numbers."""
    if chain_type == "heavy":
        positions = ahip.heavy_allowed_positions
    elif chain_type == "light":
        positions = ahip.light_allowed_positions
    else:
        raise ValueError("Unrecognized chain type passed.")

    return dict(enumerate(positions))



"""Provides a tool for humanizing input sequences according to
a preset formula."""
import copy
import numpy as np
from .sequence_scoring_tool import SequenceScoringTool
from .scoring_constants import allowed_imgt_pos as ahip


class HumanizationTool():
    """Tool for humanizing sequences."""

    def __init__(self):
        """Class constructor."""

        self.score_tool = SequenceScoringTool(offer_classifier_option = False)

        position_dict = self.score_tool.position_dict

        self.cdr_mask = {k:np.ones((len(v)), bool) for
                (k,v) in position_dict.items()}

        for chain_type, excluded_pos in ahip.CDR_EXCLUDED.items():
            for pos in excluded_pos:
                self.cdr_mask[chain_type][position_dict[chain_type][pos]] = False



    def suggest_mutations(self, seq:str, excluded_positions:list = [],
            s_thresh:float = 1.25):
        """Takes an input sequence, scores it per position,
        uses the nclusters closest clusters to determine which
        modification would be most likely to have an impact,
        suggest mutations and report both the mutations and the
        new score. CDRs are excluded, together with user-specified
        excluded positions.

        Args:
            seq (str): The sequence to update.
            s_thresh (float): The maximum percentage by which
                the score can shift before backmutation stops.
                Smaller values (closer to 1) will prioritize
                increasing the score over preserving the original
                sequence. Larger values will prioritize preserving
                the original sequence.
            excluded_positions (list): A list of strings (IMGT position numbers)
                indicating positions which should not be changed. This enables
                the user to mask key residues, Vernier zones etc if so
                desired.

        Returns:
            initial_score (float): The score of the sequence pre-modification.
            final_scores (float): The scores of the sequence after each mutation
                is adopted (in sequential order).
            mutations (list): The suggested mutations in AA_position_newAA format,
                where position is the IMGT number for the mutation position. These
                are in sequential order (the same as final_scores).
            updated_seq (list): The updated sequences after each mutation with all
                gaps removed. (This may be a different length from the
                input sequence if the suggested mutation is a deletion or
                an insertion).
        """
        original_seq, chain_name, original_arr, _, _, backmap = \
                self.score_tool._full_prep_sequence(seq)
        if chain_name not in ["H", "L"]:
            raise ValueError("The sequence provided does not recognizably "
                    "belong as a heavy or light chain.")

        mixmodel = self.score_tool.models["human"][chain_name]

        updated_arr = original_arr.copy()
        _, _, best_cluster, _ = self.score_tool.get_closest_clusters(seq, 1, False)

        best_cluster = np.log(best_cluster[0,...].clip(min=1e-16))
        best_aas = best_cluster.argmax(axis=1)
        mask = self.cdr_mask[chain_name].copy()
        for position in excluded_positions:
            mask[self.score_tool.position_dict[chain_name][position]] = False

        updated_arr[0,mask] = best_aas[mask]
        starting_score = mixmodel.score(updated_arr, n_threads = 1)[0]
        updated_score = copy.copy(starting_score)

        while updated_score > s_thresh * starting_score:
            score_shifts = best_cluster[np.arange(best_cluster.shape[0]), updated_arr.flatten()] - \
                    best_cluster[np.arange(best_cluster.shape[0]), original_arr.flatten()]
            score_shifts[~mask] = np.inf
            score_shifts[updated_arr[0,:]==original_arr[0,:]] = np.inf
            weakest_position = score_shifts.argmin()
            proposal_arr = updated_arr.copy()
            if proposal_arr[0,weakest_position] == original_arr[0,weakest_position]:
                break
            proposal_arr[0,weakest_position] = original_arr[0,weakest_position]
            updated_score = mixmodel.score(proposal_arr, n_threads = 1)[0]
            if updated_score > s_thresh * starting_score:
                updated_arr = proposal_arr
            else:
                break

        updated_seq = [self.score_tool.aa_list[aa] for aa
                in updated_arr[0,...].tolist()]

        all_mutations = []

        for i, (original_aa, new_aa) in enumerate(zip(original_seq, updated_seq)):
            if new_aa != original_aa:
                if i in backmap:
                    mutation_description = (f"{original_aa}_"
                        f"{backmap[i]+1}_{new_aa}")
                else:
                    counter = i
                    while counter not in backmap and counter > 0:
                        counter -= 1
                    if counter == 0:
                        mutation_description = (f"{original_aa}_"
                            f"{0}_{new_aa}")
                    else:
                        mutation_description = (f"{original_aa}_"
                            f"{backmap[counter]}insert_{new_aa}")
                all_mutations.append(mutation_description)

        score = mixmodel.score(updated_arr, n_threads = 1) - \
                        self.score_tool.score_adjustments[chain_name]
        return score[0], all_mutations, "".join(updated_seq).replace("-", "")



"""Tests basic functionality for the VJGeneTool."""
import os
import copy
import gzip
import unittest
import numpy as np
import blosum as bl
from antpack import VJGeneTool, SingleChainAnnotator

class TestVJGeneTool(unittest.TestCase):


    def test_error_checking(self):
        """Check that sequences which have known issues are flagged
        as such, and that deliberately invalid inputs are recognized."""
        # Pass dummy sequences with errors.
        vj_tool = VJGeneTool()
        vgene, jgene, vident, jident = vj_tool.assign_vj_genes(
                (["1", "2", "3"], 0, "H", ""),
                "AYAYAYA", "human", "identity")
        self.assertTrue(vident==0)
        self.assertTrue(jident==0)
        self.assertTrue(vgene=="")
        self.assertTrue(jgene=="")

        with self.assertRaises(RuntimeError):
            vj_tool.assign_vj_genes(
                (["1", "2", "3"], 0, "H", ""),
                "AYA", "platypus", "identity")

        with self.assertRaises(RuntimeError):
            vj_tool.assign_vj_genes(
                (["1", "2", "3"], 0, "H", ""),
                "AYA", "human", "robot")



    def test_gene_retrieval(self):
        """Make sure that we can retrieve genes and gene families
        using the appropriate methods."""
        vj_tool = VJGeneTool()
        seq = vj_tool.get_vj_gene_sequence("IGHV2-26*01", "human")
        self.assertTrue(seq == "QVTLKESGP-VLVKPTETLTLTCTVSGFSLS--NARMGVSWIRQPPGKALEWLAHIFSN---DEKSYSTSLK-SRLTISKDTSKSQVVLTMTNMDPVDTATYYCARI---------------------")

        seq = vj_tool.get_vj_gene_sequence("cow", "human")
        self.assertTrue(seq == "")


    def test_percent_ident_calc(self):
        """Double checks the percent identity calculation
        done by the cpp extension using a simple stupid
        Python version."""
        vj_tool = VJGeneTool()

        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            test_seqs = [line.strip().split(",")[0] for line in fhandle]

        germline_seqs, germline_names = vj_tool.get_seq_lists()

        sc_annotator = SingleChainAnnotator()

        os.chdir(current_dir)
        for seq in test_seqs[:100]:
            alignment = sc_annotator.analyze_seq(seq)
            chain = alignment[2]
            fmt_seq = prep_sequence(seq, alignment)

            vpred, jpred, videntity, jidentity = vj_tool.assign_vj_genes(alignment,
                    seq, "human", "identity")

            gpreds, gidentities = (vpred, jpred), (videntity, jidentity)

            for recep, gpred, id_pred in zip([f"IG{chain}V", f"IG{chain}J"],
                    gpreds, gidentities):

                best_pid, matchnum = 0, 0
                for i, template in enumerate(germline_seqs[f"human_{recep}"]):
                    matches, ntot = 0, 0

                    for qletter, tletter in zip(fmt_seq, template):
                        if tletter == "-":
                            continue
                        ntot += 1
                        if qletter == tletter:
                            matches += 1

                    true_pid = float(matches) / float(ntot)
                    if true_pid > best_pid:
                        matchnum = i
                        best_pid = copy.deepcopy(true_pid)

                gtrue = germline_names[f"human_{recep}"][matchnum]
                self.assertTrue(gtrue in gpred)
                self.assertTrue(np.allclose(id_pred, best_pid))



    def test_evalue_calc(self):
        """Double checks the evalue calculation
        done by the cpp extension using a
        simple stupid Python version."""
        vj_tool = VJGeneTool()

        blosum_matrix = bl.BLOSUM(62)

        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            test_seqs = [line.strip().split(",")[0] for line in fhandle]

        germline_seqs, germline_names = vj_tool.get_seq_lists()

        sc_annotator = SingleChainAnnotator()

        os.chdir(current_dir)
        for seq in test_seqs[:100]:
            alignment = sc_annotator.analyze_seq(seq)
            chain = alignment[2]
            fmt_seq = prep_sequence(seq, alignment)

            vpred, jpred, videntity, jidentity = vj_tool.assign_vj_genes(alignment,
                    seq, "human", "evalue")

            gpreds, gidentities = (vpred, jpred), (videntity, jidentity)

            for recep, gpred, id_pred in zip([f"IG{chain}V", f"IG{chain}J"],
                    gpreds, gidentities):

                best_pid, matchnum = -1e10, 0
                for i, template in enumerate(germline_seqs[f"human_{recep}"]):
                    score = 0

                    for qletter, tletter in zip(fmt_seq, template):
                        if tletter == "-":
                            continue
                        if qletter == "-":
                            continue

                        score += blosum_matrix[qletter][tletter]


                    if score > best_pid:
                        matchnum = i
                        best_pid = copy.deepcopy(score)

                gtrue = germline_names[f"human_{recep}"][matchnum]
                self.assertTrue(gtrue in gpred)
                self.assertTrue(np.allclose(id_pred, best_pid))


    def test_vj_assignment(self):
        """Checks vj assignments against those done by other
        tools to ensure that they are usually the same."""
        vj_tool = VJGeneTool()
        sc_annotator = SingleChainAnnotator()

        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        vhmatches, vklmatches, vhtests, vkltests = 0, 0, 0, 0
        jmatches, ntests = 0, 0

        with gzip.open("vj_gene_testing.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()

            for line in fhandle:
                seq, vgene, jgene = line.strip().split(",")
                # Eliminate problematic sequences (e.g. missing cysteine).
                alignment = sc_annotator.analyze_seq(seq)
                pid, chain, err = alignment[1], alignment[2], alignment[3]
                if pid < 0.8 or err != "":
                    continue
                pred_vgene, pred_jgene, pidv, pidj = vj_tool.assign_vj_genes(alignment,
                        seq, "human", "identity")

                if vgene in pred_vgene:
                    if chain == "H":
                        vhmatches += 1
                    else:
                        vklmatches += 1
                if jgene in pred_jgene:
                    jmatches += 1
                if chain == "H":
                    vhtests += 1
                else:
                    vkltests += 1
                ntests += 1

        print(f"On {vhtests}, vhgene, {vhmatches} success.")
        print(f"On {vkltests}, vklgene, {vklmatches} success.")
        print(f"On {ntests}, jgene, {jmatches} success.")

        self.assertTrue((vhmatches / vhtests) > 0.9)
        self.assertTrue((vklmatches / vkltests) > 0.9)
        self.assertTrue((jmatches / ntests) > 0.9)

        os.chdir(current_dir)


    def test_scheme_switching(self):
        """Checks vj assignments using different schemes to
        ensure they match."""
        imgt_tool = VJGeneTool(scheme="imgt")
        imgt_aligner = SingleChainAnnotator(scheme="imgt")

        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        for alternate_scheme in ["aho", "kabat", "aho"]:
            alternate_tool = VJGeneTool(scheme=alternate_scheme)
            alternate_aligner = SingleChainAnnotator(scheme=alternate_scheme)

            with gzip.open("vj_gene_testing.csv.gz", "rt") as fhandle:
                _ = fhandle.readline()

                for line in fhandle:
                    seq, _, _ = line.strip().split(",")
                    annotation1 = imgt_aligner.analyze_seq(seq)
                    annotation2 = alternate_aligner.analyze_seq(seq)

                    pred_vgene1, pred_jgene1, pidv1, pidj1 = imgt_tool.assign_vj_genes(annotation1,
                            seq, "human", "identity")
                    pred_vgene2, pred_jgene2, pidv2, pidj2 = alternate_tool.assign_vj_genes(
                            annotation2, seq, "human", "identity")

                    self.assertTrue(pred_vgene1==pred_vgene2)
                    self.assertTrue(pred_jgene1==pred_jgene2)
                    self.assertTrue(np.allclose(pidv1, pidv2))
                    self.assertTrue(np.allclose(pidj1, pidj2))

        os.chdir(current_dir)





def prep_sequence(sequence, alignment):
    """Converts an imgt-formatted sequence into a query for
    matching up to VJ genes."""
    numbering = alignment[0]
    std_positions = {str(i+1) for i in range(128)}

    fmt_seq = ['-' for i in range(128)]
    for letter, nb_assign in zip(sequence, numbering):
        if nb_assign in std_positions:
            fmt_seq[int(nb_assign) - 1] = letter

    return "".join(fmt_seq)



if __name__ == "__main__":
    unittest.main()



"""Contains the tools needed to generate .npy arrays for each
chain type. The .npy arrays are used to score new sequences
so that they are aligned and numbered."""
import os
import numpy as np
from Bio.Align import substitution_matrices
from ..constants import aho_default_params as aho_dp
from ..constants import imgt_default_params as imgt_dp
from ..constants import kabat_default_params as kabat_dp
from ..constants import martin_default_params as martin_dp
from ..constants import all_scheme_default_params as shared_dp


def build_consensus_files():
    fdir = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(fdir, "..", "consensus_data")
    cdir = os.getcwd()

    for scheme_name in ["kabat", "martin", "imgt", "ctermfinder", "aho"]:
        for chain_name in ["H", "K", "L"]:
            confile = f"{scheme_name.upper()}_CONSENSUS_{chain_name}.txt"
            build_scoring_files(data_path, cdir, confile,
                        chain_type = chain_name,
                        scheme = scheme_name)


def build_scoring_files(target_dir, current_dir, consensus_file,
        chain_type = "H", scheme = "kabat"):
    """Builds a scoring matrix for the amino acids at each position for each
    chain type using predefined consensus files.

    Args:
        target_dir (str): The filepath of the output directory.
        current_dir (str): The filepath of the current directory.
    """
    if chain_type not in ["H", "K", "L"]:
        raise ValueError("Currently only H, K, L chains are supported.")
    os.chdir(target_dir)
    consensus_list = load_consensus_file(consensus_file)
    if scheme == "kabat":
        save_consensus_array(consensus_list, chain_type, kabat_dp, "kabat")
    elif scheme == "martin":
        save_consensus_array(consensus_list, chain_type, martin_dp, "martin")
    elif scheme == "imgt":
        save_consensus_array(consensus_list, chain_type, imgt_dp, "imgt")
    elif scheme == "aho":
        save_consensus_array(consensus_list, chain_type, aho_dp, "aho")
    elif scheme == "ctermfinder":
        save_cterm_finder_array(consensus_list, chain_type)

    os.chdir(current_dir)



def load_consensus_file(consensus_file):
    """Loads a specified consensus file and stores it as a list
    of lists which can be converted to a scoring matrix."""
    consensus_list = []
    with open(consensus_file, "r", encoding="utf-8") as fhandle:
        position_number = 0
        for line in fhandle:
            if line.startswith("#") or line.startswith("/"):
                continue
            if len(line) <= 1:
                continue
            new_position_number = int(line.split(",")[0])
            if new_position_number != position_number + 1:
                raise ValueError(f"Consensus file {consensus_file} has incorrect formatting!")
            position_number = int(line.split(",")[0])
            observed_aas = line.strip().split(",")[1:]
            if len(observed_aas) < 1:
                raise ValueError(f"Consensus file {consensus_file} has incorrect formatting!")
            consensus_list.append(observed_aas)
    return consensus_list



def save_consensus_array(consensus_list, chain_type, constants = imgt_dp, scheme = "imgt"):
    """Converts a list of sequences for a specific chain type
    to an array with the score for each possible amino acid substitution at
    each position, including gap penalties. For IMGT (as for other numbering
    schemes), we prefer to place insertions at specific places, so we tailor
    the gap penalties to encourage this. Meanwhile, other positions are
    HIGHLY conserved, so we tailor the penalties to encourage this as well.
    IMGT numbers from 1 so we have to adjust for this."""
    allowed_amino_acids = {"A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"}
    #The list of allowed amino acids. Must be in alphabetical order, very important.
    allowed_aa_list = sorted(list(allowed_amino_acids))

    blosum = substitution_matrices.load("BLOSUM62")
    blosum_key = {letter:i for i, letter in enumerate(blosum.alphabet)}


    if chain_type.endswith("K") or chain_type.endswith("L"):
        conserved_positions = constants.light_conserved_positions
        if scheme == "aho" and chain_type.endswith("K"):
            special_positions = constants.kappa_special_positions
        else:
            special_positions = constants.light_special_positions
        cdrs = constants.light_cdrs
        npositions = constants.NUM_LIGHT

    elif chain_type.endswith("H"):
        conserved_positions = constants.heavy_conserved_positions
        special_positions = constants.heavy_special_positions
        cdrs = constants.heavy_cdrs
        npositions = constants.NUM_HEAVY

    else:
        return


    key_array = np.zeros((npositions, 23))

    for i, observed_aas in enumerate(consensus_list[:npositions]):
        position = i + 1

        #Choose the gap penalty
        #for template (column 20) and for query (column 21) of key array.
        if position in conserved_positions:
            key_array[i,21:] = shared_dp.HIGHLY_CONSERVED_GAP_PENALTY
        elif position in special_positions:
            key_array[i,21] = special_positions[position][0]
            key_array[i,22:] = special_positions[position][1]
        elif position in cdrs:
            key_array[i,21:] = cdrs[position]
        elif position in shared_dp.n_terminal_gap_positions:
            key_array[i,21] = shared_dp.n_terminal_gap_positions[position][0]
            key_array[i,22:] = shared_dp.n_terminal_gap_positions[position][1]
        else:
            key_array[i,21] = shared_dp.DEFAULT_QUERY_GAP_PENALTY
            key_array[i,22] = shared_dp.DEFAULT_TEMPLATE_GAP_PENALTY

        #Next, fill in the scores for other amino acid substitutions. If a conserved
        #residue, use the ones we specify here. Otherwise, use the best possible
        #score given the amino acids observed in the alignments. If the only
        #thing observed in the alignments is gaps, no penalty is applied.
        for j, letter in enumerate(allowed_aa_list):
            letter_blosum_idx = blosum_key[letter]
            if position in conserved_positions:
                if letter == conserved_positions[position]:
                    key_array[i,j] = shared_dp.HIGHLY_CONSERVED_BONUS
            else:
                key_array[i,j] = max([blosum[letter_blosum_idx, blosum_key[k]] if
                    k != "-" else 0 for k in observed_aas])

        #Finally, fill in the score for X. X should be assumed to be the best
        #possible amino acid that could be present at that position.
        key_array[i,20] = key_array[i,:20].max()
        if position in conserved_positions:
            key_array[i,20] = shared_dp.HIGHLY_CONSERVED_BONUS

    np.save(f"{scheme.upper()}_CONSENSUS_{chain_type}.npy", key_array)



def save_cterm_finder_array(consensus_list, chain_type):
    """Converts a list of sequences for a specific chain type
    to an array with the score for each possible amino acid substitution at
    each position SPECIFIC to the specialized c-terminal finder. Each
    covers an array of shape """
    allowed_amino_acids = {"A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"}
    #The list of allowed amino acids. Must be in alphabetical order, very important.
    allowed_aa_list = sorted(list(allowed_amino_acids))

    blosum = substitution_matrices.load("BLOSUM62")
    blosum_key = {letter:i for i, letter in enumerate(blosum.alphabet)}

    key_array = np.zeros((len(consensus_list), 21))

    for i, observed_aas in enumerate(consensus_list):

        # Fill in the scores for amino acid substitutions. For this (unlike
        # the aligner), we do not provide a large bonus for matching at
        # conserved positions, and we do not need to consider gaps.If the only
        # thing observed in the alignments is gaps, no penalty is applied.
        for j, letter in enumerate(allowed_aa_list):
            letter_blosum_idx = blosum_key[letter]
            key_array[i,j] = max([blosum[letter_blosum_idx, blosum_key[k]] if
                    k != "-" else 0 for k in observed_aas])

        #Finally, fill in the score for X. X should be assumed to be the best
        #possible amino acid that could be present at that position.
        key_array[i,20] = key_array[i,:20].max()

    np.save(f"CTERMFINDER_CONSENSUS_{chain_type}.npy", key_array)



seq = "QCKTSPLSNWHTFLFEYVHLQQSGAELMKPGASVKISCKASGYTFITYWIEWVKQRPGHGLEWIGDILPGSGSTNYNENFKGKATFTADSSSNTAYMQLSSLTSEDSAVYYCARSGYYGNSGFAYWGQGTLVTVSAHVSRTKCVADPAYSMIMDHWIIFVDDMTSELVLEVMVHYVWLRDIVMTQSPSSLTVTAGEKVTMSCKSSQSLLSSGNQKNYLTWYQQIPGQPPKLLIYWASTRESGVPDRFTGSGSGTDFTLTINSVQAEDLAVYYCQNDYTYPLTFGAGTKLELKVYFLEDMSVENQM"
from antpack import PairedChainAnnotator
pc = PairedChainAnnotator()
res = pc.analyze_seq(seq)
import pdb
pdb.set_trace()



"""Defines important defaults for the AHO numbering scheme."""
import copy

#The expected length of the heavy and light chains
NUM_HEAVY = 149
NUM_LIGHT = 149

############################The following are alignment parameters that
#were selected to try to ensure a good alignment. Making large changes
#to these WILL change the results, possibly substantially.

#These are positions that are essentially mandatory. A sequence that deviates
#from one of these is an alignment issue or has a very large deletion. Note that
#AHO defines position 90 as highly conserved in the sense that a hydrophobic
#amino acid is always observed here, but "hydrophobic amino acid" is not defined
#and a variety of (hydrophobic) aas are observed at that position.
heavy_conserved_positions = {23:"C", 43:"W", 106:"C", 139:"W", 140:"G", 142:"G"}
light_conserved_positions = {23:"C", 43:"W", 106:"C", 139:"F", 140:"G", 142:"G"}


#These are positions where either A) a blank in the query sequence is very common or
#B) insertions are very common. We want special template and query gap penalties for
#these positions. One or two of these are chain dependent. The first value is
#the query gap column, the second is the template gap column. Aho unlike other
#schemes has a separate special positions dictionary for kappa.
heavy_special_positions = {8:[-1,-12], 27:[-3.7,-12], 28:[-1.0,-12],
        36:[-1.1,-1.1], 63:[-1.1,-1.1],
        74:[-10.,-12.], 75:[-10.,-12.],
        85:[-1.5,-12.], 86:[-1.,-12.],
        87:[-8., -12.], 123:[-1.1,-1.1]}

light_special_positions = {8:[-1,-12], 27:[-3.8,-12.],
        28:[-0.2,-12.], 29:[-3.9,-12.],
        36:[-1.1,-1.1], 63:[-1.1,-1.1],
        74:[-10.,-12.], 75:[-10.,-12.],
        85:[-1.5,-12.], 86:[-1.1,-12.],
        87:[-8., -12.], 123:[-1.1,-1.1]}

kappa_special_positions = {8:[-1,-12], 27:[-0.9,-12.], 28:[-0.2,-12.],
        36:[-1.1,-1.1], 63:[-1.1,-1.1],
        74:[-10.,-12.], 75:[-10.,-12.],
        85:[-1.5,-12.], 86:[-1.,-12.],
        87:[-8., -12.], 123:[-1.05,-1.05]}

#Aho-defined CDRs. We have a specially defined gap penalty for each to ensure they are filled
#in the correct order. For Aho, light and heavy cdrs have the same definition. Notice that
#positions here do not exactly correspond to the Aho CDR definition, but rather to the region
#where insertions / deletions are likely.
heavy_cdrs = {26:-5.5, 30:-6., 31:-5.0, 32:-4.5, 33:-4.0, 34:-3.5, 35:-3.0, 36:-2.5, 37:-3.1,
        38:-3.6, 39:-4.1, 40:-4.6,
        # CDR 2
        58:-5., 59:-4.5, 60:-4., 61:-3.5, 62:-3., 63:-2.6,
        64: -3.1, 65:-3.6, 66:-4.1, 67:-4.6, 68:-5.1,
        #CDR 3
        109:-5.1, 110:-4.8, 111:-4.5, 112:-4.2, 113:-3.9, 114:-3.6, 115:-3.3, 116:-3.0,
        117:-2.7, 118:-2.4, 119:-2.1, 120:-1.8, 121:-1.5, 122:-1.2, 123:-1, 124:-1.1,
        125:-1.4, 126:-1.7, 127:-2.0, 128:-2.3, 129:-2.6, 130:-2.9, 131:-3.2, 132:-3.5,
        133:-3.8, 134:-4.1, 135:-4.4, 136:-4.7, 137:-5 }
light_cdrs = copy.copy(heavy_cdrs)



"""Contains the LiabilitySearchTool, which searches input
sequences for certain motifs known to correspond to
developability liabilities."""
import re
from ..numbering_tools import SingleChainAnnotator
from .scoring_constants import allowed_imgt_pos as ahip


class LiabilitySearchTool:
    """This class provides a simple toolkit for searching
    for some common motifs which may correspond to
    possible developability liabilities.

    Attributes:
        filters (list): A list of tuples. The first element of each tuple
            is the compiled regex filter to search for. The second element is
            a bool indicating whether the filter should only be applied to CDRs
            or to the whole variable sequence. The third element is a string
            describing the liability to which the filter corresponds.
        aligner (SingleChainAnnotator): A tool for aligning input chains.
        cdr_sets (dict): A dictionary mapping chain name to a set
            of IMGT-numbered positions present in the CDRs of that chain.
    """

    def __init__(self):
        """Class constructor."""
        self.search_filters = [
                (re.compile(r"N[^P][ST]"), False, "N-glycosylation"),
                (re.compile(r"N[GS]"), True, "Deamidation (elevated risk)"),
                (re.compile(r"D[DGHST]"), True, "Isomerization (elevated risk)"),
                (re.compile(r"N[AHNT]"), True, "Deamidation (moderate risk)"),
                (re.compile(r"[ND]P"), True, "pH-dependent hydrolysis (moderate risk)"),
                (re.compile(r"TS"), True, "pH-dependent hydrolysis risk"),
                (re.compile(r"[MW]"), True, "Methionine / Tryptophan oxidation moderate risk"),
                (re.compile(r"[STK]N"), True, "Deamidation (low risk)"),
                ]

        self.aligner = SingleChainAnnotator(chains = ["H", "K", "L"],
                    scheme = "imgt")

        self.cdr_sets = {"H": set(ahip.imgt_cdrs["H"]["1"] + ahip.imgt_cdrs["H"]["2"] +
                        ahip.imgt_cdrs["H"]["3"]),
                        "L": set(ahip.imgt_cdrs["L"]["1"] + ahip.imgt_cdrs["L"]["2"] +
                            ahip.imgt_cdrs["L"]["3"]),
                        "K": set(ahip.imgt_cdrs["L"]["1"] + ahip.imgt_cdrs["L"]["2"] +
                            ahip.imgt_cdrs["L"]["3"])
                        }



    def analyze_seq(self, sequence):
        """Runs all the stored filters against the input sequence
        and determines which if any are matches and what type of
        liabilities correspond.

        Args:
            sequence (str): A string. May be either lower or upper case but
                must contain only valid amino acids.

        Returns:
            liabilities (list): A list of tuples. The first element of each
                tuple is a 2-tuple of (starting position, ending position)
                numbered with the start of the sequence as 0, indicating the
                start and end of the liability. The second element of each tuple
                is a string describing the type of liability found. If the list
                is empty, no liabilities were found. If sequence numbering fails,
                the list contains a single tuple indicating the cause of the failure.
                (This can occur if the expected cysteines in the chain are not present
                at the expected positions, which is also a liability.)
        """
        numbering, _, chain_name, err = self.aligner.analyze_seq(sequence)

        # An alignment error may (often) indicate a liability. Return it with no
        # further analysis.
        if err != "":
            return [((0,0), err)]

        # Set up a mask indicating for each aa whether it is or is not part of a
        # CDR and of the variable region.
        cdr_mask = [k in self.cdr_sets[chain_name] for k in numbering]
        variable_mask = [True for k in numbering]
        for i, ntag in enumerate(numbering):
            if ntag != "-":
                break
            variable_mask[i] = False

        for i, ntag in reversed(list(enumerate(numbering))):
            if ntag != "-":
                break
            variable_mask[i] = False

        liabilities = []

        for search_filter, cdr_only, description in self.search_filters:
            for match_st in search_filter.finditer(sequence):
                span = match_st.span()
                if span[0] == span[1]:
                    continue
                if sum(variable_mask[span[0]:span[1]]) == 0:
                    continue
                if cdr_only:
                    if sum(cdr_mask[span[0]:span[1]]) == 0:
                        continue
                liabilities.append( (span, description) )

        # A last liability is the presence of cysteines OUTSIDE of
        # the standard positions. We check for this here.
        for i, (aa, imgt_pos) in enumerate(zip(sequence, numbering)):
            if aa == "C":
                if imgt_pos not in ["23", "104"]:
                    liabilities.append( ((i, i+1), "Unusual cysteine") )

        return liabilities



"""Tests basic functionality for the antpack cli. Note that this
relies on the functionality of the SingleChainAnnotator,
PairedChainAnnotator and VJGeneTool classes, so if those tests
do not pass, this one should not either."""
import os
import subprocess
import random
import gzip
import numpy as np
import unittest
from Bio import SeqIO
from antpack import PairedChainAnnotator, SingleChainAnnotator, VJGeneTool



class TestAntPackCLI(unittest.TestCase):




    def test_paired_chain_csv(self):
        """Take test heavy and light chains and pair them,
        then use the CLI to do basic analysis written to a
        csv file."""
        merged_seqs, heavy_pc_annotations, light_pc_annotations, pc_aligner = get_paired_seqs()

        heavy_code_map = create_manual_mapping(heavy_pc_annotations, pc_aligner)
        light_code_map = create_manual_mapping(light_pc_annotations, pc_aligner)

        # Create a temporary fasta input file.
        with open("input_fasta_data.fasta", "w+", encoding="utf-8") as fhandle:
            for i, merged_seq in enumerate(merged_seqs):
                fhandle.write(f">this is sequence # {i}\n{merged_seq}\n")


        _ = subprocess.run(["AntPack", "input_fasta_data.fasta",
                "output_data", "imgt", "--paired"])

        self.assertTrue("output_data_heavy.csv" in os.listdir())
        self.assertTrue("output_data_light.csv" in os.listdir())

        with open("output_data_heavy.csv", "r",
                encoding="utf-8") as fhandle:
            _ = fhandle.readline()
            for i, line in enumerate(fhandle):
                elements = line.strip().split(",")
                self.assertTrue(elements[0] == f"this is sequence # {i}")
                self.assertTrue("".join(elements[2:-1])==map_sequence(merged_seqs[i],
                    heavy_pc_annotations[i], heavy_code_map))
                self.assertTrue(elements[-1]==heavy_pc_annotations[i][-1])

            self.assertTrue((i+1)==len(heavy_pc_annotations))

        with open("output_data_light.csv", "r",
                encoding="utf-8") as fhandle:
            temp = fhandle.readline()
            for i, line in enumerate(fhandle):
                elements = line.strip().split(",")
                self.assertTrue(elements[0] == f"this is sequence # {i}")
                self.assertTrue("".join(elements[2:-1])==map_sequence(merged_seqs[i],
                    light_pc_annotations[i], light_code_map))
                self.assertTrue(elements[-1]==light_pc_annotations[i][-1])

            self.assertTrue((i+1)==len(light_pc_annotations))

        os.remove("input_fasta_data.fasta")
        os.remove("output_data_heavy.csv")
        os.remove("output_data_light.csv")




    def test_paired_chain_vj(self):
        """Take test heavy and light chains and pair them,
        then use the CLI to do basic analysis written to a
        csv file with human VJ gene assignment."""
        merged_seqs, heavy_pc_annotations, light_pc_annotations, pc_aligner = get_paired_seqs()
        vj_tool = VJGeneTool(database="imgt", scheme="imgt")

        heavy_code_map = create_manual_mapping(heavy_pc_annotations, pc_aligner)
        light_code_map = create_manual_mapping(light_pc_annotations, pc_aligner)

        # Create a temporary fasta input file.
        with open("input_fasta_data.fasta", "w+", encoding="utf-8") as fhandle:
            for i, merged_seq in enumerate(merged_seqs):
                fhandle.write(f">this is sequence # {i}\n{merged_seq}\n")


        _ = subprocess.run(["AntPack", "input_fasta_data.fasta", "output_data",
            "imgt", "--paired", "--vj", "human", "identity"])

        self.assertTrue("output_data_heavy.csv" in os.listdir())
        self.assertTrue("output_data_light.csv" in os.listdir())

        with open("output_data_heavy.csv", "r",
                encoding="utf-8") as fhandle:
            _ = fhandle.readline()

            for i, line in enumerate(fhandle):
                elements = line.strip().split(",")
                self.assertTrue(elements[0] == f"this is sequence # {i}")
                self.assertTrue(elements[2] == "human")
                self.assertTrue(elements[3] == "identity")
                v_gene, j_gene, vident, jident = vj_tool.assign_vj_genes(heavy_pc_annotations[i],
                        merged_seqs[i], "human", "identity")
                self.assertTrue(elements[4]==v_gene)
                self.assertTrue(np.allclose(float(elements[5]), vident))
                self.assertTrue(elements[6]==j_gene)
                self.assertTrue(np.allclose(float(elements[7]), jident))
                self.assertTrue("".join(elements[8:-1])==map_sequence(merged_seqs[i],
                    heavy_pc_annotations[i], heavy_code_map))
                self.assertTrue(elements[-1]==heavy_pc_annotations[i][-1])

            self.assertTrue((i+1)==len(heavy_pc_annotations))

        with open("output_data_light.csv", "r",
                encoding="utf-8") as fhandle:
            _ = fhandle.readline()
            for i, line in enumerate(fhandle):
                elements = line.strip().split(",")
                self.assertTrue(elements[0] == f"this is sequence # {i}")
                self.assertTrue(elements[2] == "human")
                self.assertTrue(elements[3] == "identity")
                v_gene, j_gene, vident, jident = vj_tool.assign_vj_genes(light_pc_annotations[i],
                        merged_seqs[i], "human", "identity")
                self.assertTrue(elements[4]==v_gene)
                self.assertTrue(np.allclose(float(elements[5]), vident))
                self.assertTrue(elements[6]==j_gene)
                self.assertTrue(np.allclose(float(elements[7]), jident))
                self.assertTrue("".join(elements[8:-1])==map_sequence(merged_seqs[i],
                    light_pc_annotations[i], light_code_map))
                self.assertTrue(elements[-1]==light_pc_annotations[i][-1])

            self.assertTrue((i+1)==len(light_pc_annotations))

        os.remove("input_fasta_data.fasta")
        os.remove("output_data_heavy.csv")
        os.remove("output_data_light.csv")






    def test_single_chain_csv(self):
        """Take test heavy and light chains,
        then use the CLI to do basic analysis and write
        to fasta, using single chain containing sequences
        as input only."""
        seqs, annotations, sc_aligner = get_unpaired_seqs()
        heavy_code_map = create_manual_mapping([a for a in annotations if
            a[2] == 'H'], sc_aligner)
        light_code_map = create_manual_mapping([a for a in annotations if
            a[2] != 'H'], sc_aligner)

        # Create a temporary fasta input file.
        with open("input_fasta_data.fasta", "w+", encoding="utf-8") as fhandle:
            for i, seq in enumerate(seqs):
                fhandle.write(f">this is sequence # {i}\n{seq}\n")

        _ = subprocess.run(["AntPack", "input_fasta_data.fasta", "output_data",
            "imgt"])

        self.assertTrue("output_data_heavy.csv" in os.listdir())
        self.assertTrue("output_data_light.csv" in os.listdir())

        ntot = 0
        for chain_type, code_map in zip(["heavy", "light"], [heavy_code_map,
            light_code_map]):
            with open(f"output_data_{chain_type}.csv", "r",
                    encoding="utf-8") as fhandle:
                _ = fhandle.readline()
                for i, line in enumerate(fhandle):
                    ntot += 1
                    elements = line.strip().split(",")
                    code = int(elements[0].split("this is sequence # ")[1])
                    self.assertTrue(annotations[code][-1] == elements[-1])
                    self.assertTrue("".join(elements[2:-1])==map_sequence(seqs[code],
                        annotations[code], code_map))

        self.assertTrue(ntot==len(annotations))

        os.remove("input_fasta_data.fasta")
        os.remove("output_data_heavy.csv")
        os.remove("output_data_light.csv")



    def test_single_chain_vj(self):
        """Take test heavy and light chains,
        then use the CLI to do basic analysis plus vj assignment
        and write to fasta, using single chain containing sequences
        as input only."""
        seqs, annotations, sc_aligner = get_unpaired_seqs()
        vj_tool = VJGeneTool(scheme="imgt", database="imgt")

        heavy_code_map = create_manual_mapping([a for a in annotations if
            a[2] == 'H'], sc_aligner)
        light_code_map = create_manual_mapping([a for a in annotations if
            a[2] != 'H'], sc_aligner)

        # Create a temporary fasta input file.
        with open("input_fasta_data.fasta", "w+", encoding="utf-8") as fhandle:
            for i, seq in enumerate(seqs):
                fhandle.write(f">this is sequence # {i}\n{seq}\n")

        _ = subprocess.run(["AntPack", "input_fasta_data.fasta", "output_data",
            "imgt", "--vj", "human", "identity"])

        self.assertTrue("output_data_heavy.csv" in os.listdir())
        self.assertTrue("output_data_light.csv" in os.listdir())

        ntot = 0
        for chain_type, code_map in zip(["heavy", "light"], [heavy_code_map,
            light_code_map]):
            with open(f"output_data_{chain_type}.csv", "r",
                    encoding="utf-8") as fhandle:
                _ = fhandle.readline()
                for i, line in enumerate(fhandle):
                    ntot += 1
                    elements = line.strip().split(",")
                    code = int(elements[0].split("this is sequence # ")[1])
                    self.assertTrue(elements[2] == "human")
                    self.assertTrue(elements[3] == "identity")
                    v_gene, j_gene, vident, jident = vj_tool.assign_vj_genes(annotations[code],
                            seqs[code], "human", "identity")
                    self.assertTrue(elements[4]==v_gene)
                    self.assertTrue(np.allclose(float(elements[5]), vident))
                    self.assertTrue(elements[6]==j_gene)
                    self.assertTrue(np.allclose(float(elements[7]), jident))
                    self.assertTrue(annotations[code][-1] == elements[-1])
                    self.assertTrue("".join(elements[8:-1])==map_sequence(seqs[code],
                        annotations[code], code_map))

        self.assertTrue(ntot==len(annotations))

        os.remove("input_fasta_data.fasta")
        os.remove("output_data_heavy.csv")
        os.remove("output_data_light.csv")



def create_manual_mapping(annotation_list, pc_tool,
        scheme = "imgt", add_unobserved_positions = True):
    """Manually creates a dictionary mapping codes to positions."""
    code_set = set()

    for annotation in annotation_list:
        code_set.update(annotation[0])

    if add_unobserved_positions:
        if scheme == "imgt":
            for i in range(1, 129):
                code_set.add(str(i))

    code_set = list(code_set)
    sorted_codes = pc_tool.sort_position_codes(code_set)
    return {k:i for i, k in enumerate(sorted_codes)}



def map_sequence(sequence, annotation, code_dict):
    """Maps an input sequence to a standard length vector to
    create an MSA."""
    output_seq = ["-" for k in code_dict]
    for letter, code in zip(sequence, annotation[0]):
        if code == "-":
            continue
        output_seq[code_dict[code]] = letter
    return "".join(output_seq)


def get_paired_seqs():
    """Get a set of paired chain sequences for testing."""
    project_path = os.path.abspath(os.path.dirname(__file__))
    current_dir = os.getcwd()
    os.chdir(os.path.join(project_path, "test_data"))
    with gzip.open("test_data.csv.gz", "rt") as fhandle:
        _ = fhandle.readline()
        seqs = [line.strip().split(",")[0] for line in fhandle]
    os.chdir(current_dir)

    sc_aligner = SingleChainAnnotator(scheme="imgt")
    pc_aligner = PairedChainAnnotator(scheme="imgt")

    alignments = [sc_aligner.analyze_seq(seq) for seq in seqs]

    heavy_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] == "H"]
    light_chains = [(seq, a) for a, seq in zip(alignments, seqs)
                if a[2] != "H"]

    random.seed(0)

    # Join the chains using a very arbitrary linker.
    merged_seqs = ["".join( [light_chain[0], "YYYYGGGGSSSS", heavy_chain[0]] )
                for (heavy_chain, light_chain) in zip(heavy_chains, light_chains)]

    heavy_pc_annotations, light_pc_annotations = [], []

    for merged_seq in merged_seqs:
        h_annot, l_annot = pc_aligner.analyze_seq(merged_seq)
        heavy_pc_annotations.append(h_annot)
        light_pc_annotations.append(l_annot)

    return merged_seqs, heavy_pc_annotations, light_pc_annotations, pc_aligner



def get_unpaired_seqs():
    """Get a set of paired chain sequences for testing."""
    project_path = os.path.abspath(os.path.dirname(__file__))
    current_dir = os.getcwd()
    os.chdir(os.path.join(project_path, "test_data"))
    with gzip.open("test_data.csv.gz", "rt") as fhandle:
        _ = fhandle.readline()
        seqs = [line.strip().split(",")[0] for line in fhandle]
    os.chdir(current_dir)

    sc_aligner = SingleChainAnnotator(scheme="imgt")

    annotations = [sc_aligner.analyze_seq(seq) for seq in seqs]

    return seqs, annotations, sc_aligner


if __name__ == "__main__":
    unittest.main()



"""Tests basic functionality for prefiltering routines."""
import os
import unittest
import gzip
import numpy as np
from antpack.numbering_tools.cterm_finder import PyCTermFinder
from antpack.numbering_tools.cterm_finder import _load_nterm_kmers

KMER_WIDTH = 9
KMER_WINDOW = 14


class TestPrefiltering(unittest.TestCase):


    def test_pycterm_routine(self):
        """Compare scoring for N and C terminals with the
        wrapped C++ code with an inefficient but simple Python
        routine to ensure results match."""
        project_path = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(os.path.join(project_path, "test_data"))

        with gzip.open("test_data.csv.gz", "rt") as fhandle:
            _ = fhandle.readline()
            seqs = [line.strip().split(",")[0] for line
                    in fhandle]

        os.chdir(current_dir)

        aa_to_idx = ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N",
            "P", "Q", "R", "S", "T", "V", "W", "Y"]
        aa_to_idx = {a:i for i,a in enumerate(aa_to_idx)}

        pyct_finder = PyCTermFinder()
        kmer_dict = _load_nterm_kmers()
        cterm_to_aa = np.zeros((10,21,3))

        for k, chain in enumerate(["H", "K", "L"]):
            cterm_to_aa[:,:,k] = np.load(
                    os.path.join(project_path, "..", "src",
                "antpack", "numbering_tools", "consensus_data",
                f"CTERMFINDER_CONSENSUS_{chain}.npy"))

        for seq in seqs:
            nterm_gt_scores, cterm_gt_scores = \
                    np.zeros((3), dtype=np.int32), \
                    np.zeros((3))

            kmer_presence = np.zeros((len(seq), 3),
                        dtype=np.int32)
            for i, _ in enumerate(seq[:-KMER_WIDTH]):
                if seq[i:i+KMER_WIDTH] in kmer_dict:
                    kmer_presence[i,
                            kmer_dict[seq[i:i+KMER_WIDTH]]] += 1

            for i, _ in enumerate(seq[:-5]):
                cterm_scores = np.zeros((3))
                len_to_traverse = min(cterm_to_aa.shape[0],
                        len(seq) - i)
                for j in range(len_to_traverse):
                    for k in range(3):
                        cterm_scores[k] += cterm_to_aa[j,
                                aa_to_idx[seq[i+j]], k]
                for k in range(3):
                    if cterm_scores[k] > cterm_gt_scores[k]:
                        cterm_gt_scores[k] = cterm_scores[k]

            for i in range(kmer_presence.shape[0] -
                    KMER_WINDOW):
                for k in range(3):
                    score = kmer_presence[i:i+KMER_WINDOW,k].sum()
                    if score > nterm_gt_scores[k]:
                        nterm_gt_scores[k] = score

            nterm_test_scores, cterm_test_scores = \
                    np.zeros((3), dtype=np.int32), \
                    np.zeros((3))
            nterm_test_positions, cterm_test_positions = \
                    np.zeros((3), dtype=np.int32), \
                    np.zeros((3), dtype=np.int32)
            pyct_finder.pyfind_c_terminals(seq, cterm_test_scores,
                    cterm_test_positions, nterm_test_scores,
                    nterm_test_positions)
            self.assertTrue(np.allclose(nterm_gt_scores,
                nterm_test_scores))
            self.assertTrue(np.allclose(cterm_gt_scores,
                cterm_test_scores))


if __name__ == "__main__":
    unittest.main()



"""Defines important defaults for the IMGT numbering scheme."""
import copy

#The expected length of the heavy and light chains
NUM_HEAVY = 128
NUM_LIGHT = 127

############################The following are alignment parameters that
#were selected to try to ensure a good alignment. Making large changes
#to these WILL change the results, possibly substantially.

#These are positions that are essentially mandatory. A sequence that deviates
#from one of these is an alignment issue or has a very large deletion. Note that
#IMGT defines position 89 as highly conserved in the sense that a hydrophobic
#amino acid is always observed here, but "hydrophobic amino acid" is not defined
#and a variety of (hydrophobic) aas are observed at that position. Hence we
#handle position 89 separately below.
heavy_conserved_positions = {23:"C", 41:"W", 104:"C", 118:"W", 119:"G", 121:"G"}
light_conserved_positions = {23:"C", 41:"W", 104:"C", 118:"F", 119:"G", 121:"G"}


#These are positions where either A) a blank in the query sequence is very common or
#B) insertions are very common. We want special template and query gap penalties for
#these positions. One or two of these are chain dependent. The first value is
#the query gap column, the second is the template gap column.
heavy_special_positions = {10:[-1,-12], 33:[-1.2,-1.2], 61:[-1.2,-1.2],
        73:[-1,-11], 111:[-1.2,-1.2]}

light_special_positions = {10:[-1,-12],
        33:[-1.2,-1.2], 61:[-1.2,-1.2], 73:[-1,-11], 81:[-1.,-25],
        82:[-1.,-25], 111:[-1.2,-1.2]}

#IMGT-defined CDRs. We have a specially defined gap penalty for each to ensure they are filled
#in the correct order. For IMGT, light and heavy cdrs have the same definition.
heavy_cdrs = {27:-5.1, 28:-4.6, 29:-4.1, 30:-3.6, 31:-3.1, 32:-2.6, 33:-2.5, 34:-3, 35:-3.5,
        36:-4, 37:-4.5, 38:-5,
        # CDR 2
        56:-5.1, 57:-4.6, 58:-4.1, 59:-3.6, 60:-3.1, 61:-3, 62:-3.5, 63:-4, 64:-4.5, 65:-5,
        #CDR 3
        105:-5.1, 106:-4.6, 107:-4.1, 108:-3.6, 109:-3.1, 110:-2.6, 111:-1,
        112:-2.5, 113:-3, 114:-3.5, 115:-4, 116:-4.5, 117:-5}
light_cdrs = copy.copy(heavy_cdrs)



"""Contains some simple utilities for loading VJ gene databases."""
import os
import gzip



def load_vj_gene_consensus_db(current_dir, database_path, database = "imgt"):
    """Loads the current vj gene databases and returns them
    as two dictionaries mapping species name and receptor type
    to a list of gene names and to a list of gene sequences.

    Args:
        current_dir (str): The current user directory.
        project_path (str): The location of the consensus databases."""
    vj_names, vj_seqs, retrieved_dates = {}, {}, {}

    try:
        os.chdir(database_path)
        db_files = [f for f in os.listdir() if f.endswith(".fa.gz") and
                    f.split("_")[1] == database]
        for db_file in db_files:
            species = db_file.split("_")[0]
            receptor = db_file.split("_")[2]
            if species not in retrieved_dates:
                retrieved_dates[species] = {}

            retrieved_dates[species][receptor] = db_file.split(".fa.gz")[0].split("_")[-1]

            # We avoid using Biopython's SeqIO (since it introduces an additional
            # unnecessary dependency). Since we wrote the db files, we know
            # how they are formatted -- one aa seq for each gene name -- and
            # can use a very simple procedure here which is not applicable to all
            # fasta files.
            with gzip.open(db_file, "rt", encoding="utf-8") as fhandle:
                sequences, names = [], []

                for line in fhandle:
                    if line.startswith(">"):
                        description = line.strip()[1:]
                    else:
                        aa_seq = line.strip()
                        sequences.append(aa_seq)
                        names.append(description)


            key = species + "_" + receptor
            if key in vj_names:
                raise RuntimeError("Duplicate db files found for "
                            f"{species}, {receptor}.")

            vj_names[key] = names
            vj_seqs[key] = sequences

    except Exception as exc:
        os.chdir(current_dir)
        raise RuntimeError("The consensus data for the package either has been deleted or "
                    "moved or was never properly installed.") from exc

    os.chdir(current_dir)
    return vj_names, vj_seqs, retrieved_dates
